"""Module for occupancy class and calculation.

Occupancy can be loosely defined as proportion of area occupied by some target
species. Unfortunately this definition is not rigorous enough to define a
single concept. Hence definitions have proliferated in the literature, mainly
guided by study restriction and convenience.

Three main issues arise when trying to clearly define occupancy:

    1. The definition of space being occupied by some species: In some cases,
       where niches are sharply delimited by geographical features, such as
       ponds, occupancy refers to the proportion of such areas occupied by
       some target species at an instant in time. But when movement is not
       restrained by clear boundaries occupancy is harder to
       define. One approach is defining occupancy as the proportion of points
       at which individuals of the target species are observed, having
       preselected the points as being "representative". This seems to be an
       ad hoc definition for camera trap studies. Finally when movement data
       is available occupancy can be defined, more faithful to the original
       concept, as the area visited by the individuals of the target species.
       But this also introduces the following two problems:
    2. Calculation of area visited by an individual usually starts by some sort
       discretization of space, and is done by counting the number of cells of
       discretized space that individuals touch along its movement. This
       implies that there will be some dependence of the final occupancy value
       on the method of space discretization.
    3. Finally, when occupancy is calculated from movement data, there must be
       a choice of time span for which to calculate occupancy. Occupancy values
       will be sensitive to choice of time span, or could even be result in a
       useless feature if, for example, individuals of species fully roam the
       available land given enough time, even though they usually take a small
       proportion of land area at smaller time scales.

Due to the nature of the data generated in movement simulations we have chosen
to define occupancy as the result of the following calculations:

1. Select some definite time span, say 90 days of movement.
2. Use some pre defined function to obtain spatial resolution from home range
   data and discretize study site with such resolution. Symbolically::

      resolution = f(home_range, parameters)

   See :py:func:`ollin.core.utils.occupancy_resolution`

3. For each time step mark as visited all cells that have some individual of
   the target species contained in it.
4. For each cell of discretized space, calculate the proportion of times the
   cell was visited.
5. Occupancy is defined as the average of this proportion of times cells where
   visited.

We have chosen to use proportion of occupancy, or rate of occupancy, since
this value is less sensitive to changes in time span selected for occupancy
calculation. In simulated situations it will always converge to some specific
value as time span grows.

All values used in the occupancy calculation, such as the grid of visited cells
per time step, number of times a cell was visited and others, are stored within
a :py:obj:`Occupancy` object.

"""
from __future__ import division

from six.moves import xrange
import numpy as np
from numba import jit, float64

from .utils import occupancy_resolution


class Occupancy(object):
    """Occupancy information and calculation class.

    Occupancy can be calculated from movement data in the following way:

    1. Select some spatial resolution to use in space discretization.
    2. Discretize site to obtain n x m cells.
    3. Initialize an array of shape [time_steps, n, m] to zeros.
    4. Set ::

          array[t, i, j] = 1

       if some individual is in cell (i,j) at time step t.
    5. Occupancy is the average of this array.

    Attributes
    ----------
    movement : :py:obj:`ollin.MovementData`
        Movement data for which to calculate occupancy.
    steps : int
        Number of movement steps contained in movement data.
    resolution : float
        Spatial resolution (in Km) for site discretization.
    grid : array
        Array of shape [time_steps, x, y] where [x, y] is the
        size of the discretized site. Holds cell occupancy at
        each time step.
    cell_occupancy : array
        Array of shape [x, y] where [x, y] is the size of the
        discretized site. Holds cell occupancy.
    occupancy : float
        Occupancy measure.
    """

    def __init__(self, movement, resolution=None):
        """Construct Occupancy object.

        Arguments
        ---------
        movement : :py:obj:`ollin.MovementData`
            Movement data to be analized.
        resolution : float, optional
            Resolution for space discretization. If none is given,
            resolution will be calculated from home_range data
            stored in the movement data.
        """
        self.movement = movement
        self.steps = movement.steps

        if resolution is None:
            resolution = occupancy_resolution(movement.home_range)
        self.resolution = resolution

        self.grid = make_grid(self.movement, self.resolution)
        self.cell_occupancy = self.grid.sum(axis=0) / self.steps
        self.occupancy = self.cell_occupancy.mean()

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            occupancy_cmap='Blues',
            occupancy_level=0.2,
            occupancy_alpha=0.3,
            **kwargs):
        """Plot discretized space occupancy information.

        Occupancy plotting adds the following optional components to the
        plot:

        1. "occupancy":
            If present in include list, cells of discretized space will be
            shown with a color representing the rate of occupation of such
            cell. Colors will be selected using a color map. See
            :py:mod:`matplotlib.cm` to see all colormap options.
        2. "occupancy_contour":
            If present in include list, the border of the region containing all
            cells with occupancy heigher than some threshold will be plotted.

        All other components in the include list will be passed down to the
        MovementData plotting method. See :py:meth:`ollin.MovementData.plot`
        for all plot components defined at that level.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot occupancy information.
        figsize : list or tuple, optional
            Size of figure to create if no axes object was given. Defaults to
            (10, 10).
        include : list or tuple, optional
            List of components to add to the plot. Components list will be
            passed to the Site object to add the corresponding components.
        occupancy_cmap : str, optional
            Colormap to use to codify occupancy level to color. Defaults to
            'Blues'.
        occupancy_level : float, optional
            Threshold at which to draw boundary for occupancy contour.
        occupancy_alpha : float, optional
            Alpha value of cell colors.
        kwargs : dict, optional
            All other keyworded arguments will be passed to the MovementData's
            plotting method.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Return axes for further plotting.

        """
        import matplotlib.pyplot as plt  # pylint: disable=import-error
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if include is None:
            include = [
                'rectangle',
                'niche_boundary',
                'occupancy',
                'occupancy_contour']

        self.movement.plot(include=include, ax=ax, **kwargs)

        if 'occupancy' in include:
            grid = self.cell_occupancy

            range_ = self.movement.site.range
            h, w = grid.shape
            xcoord, ycoord = np.meshgrid(
                np.linspace(0, range_[0], h),
                np.linspace(0, range_[1], w))
            cm = ax.pcolormesh(
                xcoord,
                ycoord,
                grid.T,
                cmap=occupancy_cmap,
                alpha=occupancy_alpha,
                vmax=1.0,
                vmin=0.0)
            plt.colorbar(cm, ax=ax)

            if 'occupancy_contour' in include:
                mask = (grid >= occupancy_level)
                ax.contour(xcoord, ycoord, mask.T, levels=[0.5], cmap='Blues')

        return ax

    def plot_occupancy_timeseries(
            self,
            ax=None,
            figsize=(10, 10),
            color='red',
            label=''):
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        times = self.movement.times
        occupancies = [self.grid[:n].mean() for n in xrange(2, times.size)]

        ax.plot(times[2:], occupancies, color=color, label=label)
        ax.set_xlabel('Time (Days)')
        ax.set_ylabel('Occupancy')
        ax.set_ylim(0, 1)

        return ax


@jit(
    float64[:, :, :](
        float64[:, :, :],
        float64[:],
        float64),
    nopython=True)
def _make_grid(array, range, resolution):
    num_sides_x = int(np.ceil(range[0] / resolution))
    num_sides_y = int(np.ceil(range[1] / resolution))

    num, steps, _ = array.shape

    space = np.zeros((steps, num_sides_x, num_sides_y))

    denominator = np.zeros((2,))
    denominator[0] = range[0] / num_sides_x
    denominator[1] = range[1] / num_sides_y
    indices = np.floor_divide(array, denominator).astype(np.int64)

    disc_range = np.zeros((2,), dtype=np.int64)
    disc_range[0] = num_sides_x - 1
    disc_range[1] = num_sides_y - 1
    indices = np.minimum(indices, disc_range)

    for s in xrange(steps):
        for i in xrange(num):
            x, y = indices[i, s]
            space[s, x, y] = 1
    return space


def make_grid(mov, resolution):
    """Make and return grid of occupancies from movement data."""
    return _make_grid(mov.data, mov.site.range, resolution)



from six.moves import xrange
import numpy as np
from numba import jit, float64, int64

from .basemodel import MovementModel


class Model(MovementModel):
    name = 'Constant Brownian Model'
    default_parameters = {
        'movement': {},
        "density": {
            "alpha": 4.950773510732457,
            "density_exp": 0.8501718469622543,
            "hr_exp": 0.8441724458551622,
            "niche_size_exp": 0.8760285191282491
            },
        "home_range": {
            "alpha": 59.02095500748234,
            "exponent": 1.903072119815655
            },
        "velocity": {
            "alpha": 0.002223569915673946,
            "beta": 1.013044346939526
            }
    }

    def generate_movement(
            self,
            initial_positions,
            site,
            steps,
            velocity):
        range_ = site.range
        mov = self._movement(
            initial_positions,
            velocity,
            range_,
            steps)
        return mov

    @staticmethod
    @jit(
        float64[:, :, :](
            float64[:, :],
            float64,
            float64[:],
            int64),
        nopython=True)
    def _movement(
            random_positions,
            velocity,
            range_,
            steps):
        num, _ = random_positions.shape
        movement = np.zeros((num, steps, 2), dtype=float64)
        sigma = velocity / 1.2533141373155003
        rangex, rangey = range_
        directions = np.random.normal(
            0, sigma, size=(steps, num, 2))

        for k in xrange(steps):
            movement[:, k, :] = random_positions
            for j in xrange(num):
                direction = directions[k, j]
                tmp1 = (
                    random_positions[j, 0] + direction[0],
                    random_positions[j, 1] + direction[1])
                tmp2 = (tmp1[0] % (2 * rangex), tmp1[1] % (2 * rangey))

                if tmp2[0] < rangex:
                    random_positions[j, 0] = tmp2[0] % rangex
                else:
                    random_positions[j, 0] = (-tmp2[0]) % rangex

                if tmp2[1] < rangey:
                    random_positions[j, 1] = tmp2[1] % rangey
                else:
                    random_positions[j, 1] = (-tmp2[1]) % rangey
        return movement



from __future__ import print_function
from multiprocessing import Pool

import sys
import numpy as np
import ollin


RANGE = 20
TRIALS_PER_WORLD = 100
NUM_WORLDS = 10
VELOCITIES = np.linspace(0.1, 3, 10)
NICHE_SIZES = np.linspace(0.3, 0.9, 6)


class VelocityCalibrator(object):
    def __init__(
            self,
            movement_model,
            velocities=VELOCITIES,
            niche_sizes=NICHE_SIZES,
            trials_per_world=TRIALS_PER_WORLD,
            num_worlds=NUM_WORLDS,
            range=RANGE):

        self.movement_model = movement_model
        self.velocities = velocities
        self.niche_sizes = niche_sizes
        self.trials_per_world = trials_per_world
        self.num_worlds = num_worlds
        self.range = range

        self.velocity_info = self.calculate_velocity_info()

    def calculate_velocity_info(self):
        n_vel = len(self.velocities)
        n_nsz = len(self.niche_sizes)
        num = self.trials_per_world
        mov = self.movement_model

        all_info = np.zeros(
            [n_vel, n_nsz, self.num_worlds, self.trials_per_world])
        arguments = [
            Info(mov, vel, nsz, num, self.range)
            for vel in self.velocities
            for nsz in self.niche_sizes
            for k in xrange(self.num_worlds)]

        print('Simulating {} scenarios'.format(len(arguments)))
        pool = Pool()
        try:
            results = pool.map(get_single_velocity_info, arguments)
            pool.close()
            pool.join()
        except KeyboardInterrupt:
            pool.terminate()
            sys.exit()
            quit()

        arguments = [
            (i, j, k)
            for i in xrange(n_vel)
            for j in xrange(n_nsz)
            for k in xrange(self.num_worlds)]

        for (i, j, k), res in zip(arguments, results):
            all_info[i, j, k, :] = res

        return all_info

    def plot(self, cmap='Set2', figsize=(10, 10), ax=None):
        import matplotlib.pyplot as plt
        from matplotlib.cm import get_cmap

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        cmap = get_cmap(cmap)

        for n, nsz in enumerate(self.niche_sizes):
            color = cmap(float(n) / len(self.niche_sizes))
            data = self.velocity_info[:, n, :, :]
            mean = data.mean(axis=(1, 2))
            std = data.std(axis=(1, 2))

            ax.plot(
                self.velocities,
                mean,
                c=color,
                label='Niche Size: {}'.format(nsz))
            ax.fill_between(
                self.velocities,
                mean - std,
                mean + std,
                color=color,
                alpha=0.6,
                edgecolor='white')

        ax.plot(self.velocities, self.velocities, color='red', label='target')
        ax.set_title('Velocity Calibration')

        ax.set_xlabel('target velocity (Km/day)')
        ax.set_ylabel('calculated velocity (Km/day)')

        ax.set_xticks(self.velocities)
        ax.set_yticks(self.velocities)

        ax.legend()
        return ax

    def fit(self):
        from sklearn.linear_model import LinearRegression

        coefficients = np.zeros(len(self.niche_sizes))
        for num, niche_size in enumerate(self.niche_sizes):
            given_vel = []
            simulated_vel = []

            for k, vel in enumerate(self.velocities):
                vdata = self.velocity_info[k, num, :, :].ravel()
                given_vel.append(vel * np.ones_like(vdata))
                simulated_vel.append(vdata)

            given_vel = np.concatenate(given_vel, 0)
            simulated_vel = np.concatenate(simulated_vel, 0)
            model = LinearRegression(fit_intercept=False)
            model.fit(given_vel[:, None], simulated_vel)
            coefficients[num] = 1 / model.coef_[0]

        model = LinearRegression()
        model.fit(self.niche_sizes[:, None], coefficients)

        alpha = model.coef_[0]
        beta = model.intercept_

        fit = {
            'alpha': alpha,
            'beta': beta}
        return fit


class Info(object):
    __slots__ = [
        'movement_model',
        'velocity',
        'niche_size',
        'num',
        'range']

    def __init__(
            self,
            movement_model,
            velocity,
            niche_size,
            num,
            range_):
        self.movement_model = movement_model
        self.velocity = velocity
        self.niche_size = niche_size
        self.num = num
        self.range = range_


def get_single_velocity_info(info):
    site = ollin.Site.make_random(info.niche_size, range=info.range)
    mov = ollin.Movement.simulate(
        site,
        num=info.num,
        velocity=info.velocity,
        days=info.movement_model.parameters['hr_days'],
        movement_model=info.movement_model)
    analyzer = ollin.MovementAnalysis(mov)
    return analyzer.velocities.mean(axis=1)



"""Base class for all movement models."""

from abc import abstractmethod
from six import iteritems

from ..core.constants import MOVEMENT_PARAMETERS


class MovementModel(object):
    """Base class for all movement models

    Any movement model must subclass this class and implement the generate
    movement method. Also they must provide a global attribute defining the
    default parameters for any instance of the movement model.

    Attributes
    ----------
    name : str
        Name of model. Mainly used in plotting.
    default_parameters : dict
        Dictionary holding default values of any parameters associated with the
        movement model. There are some required parameters for every
        movement model. If not provided they will default to those in
        :py:const:`ollin.core.constants.MOVEMENT_PARAMETERS`.
    """
    name = None
    default_parameters = {}

    def handle_parameters(self, params):
        """Return parameter dictionary values with missing default values.

        This function searchs for all required parameters in the dictionary
        passed as argument. If not found it will default to the value specified
        in the default_parameters dictionary of the movement model class. If
        not found it will default to the one specified in the constants module.
        (see :py:const:`pycamptra.core.constants.MOVEMENT_PARAMETERS`)
        """
        if params is None:
            params = {}

        parameters = MOVEMENT_PARAMETERS.copy()

        for key, value in iteritems(self.default_parameters):
            if isinstance(value, dict):
                try:
                    parameters[key].update(value)
                except KeyError:
                    parameters[key] = value
            else:
                parameters[key] = value

        for key, value in iteritems(params):
            if isinstance(value, dict):
                try:
                    parameters[key].update(value)
                except KeyError:
                    parameters[key] = value
            else:
                parameters[key] = value

        return parameters

    def __init__(self, parameters=None):
        """Construct a movement model instance with the given parameters"""
        if parameters is None:
            parameters = {}
        self.parameters = self.handle_parameters(parameters)

    @abstractmethod
    def generate_movement(
            self,
            initial_position,
            site,
            steps,
            velocity):
        """Generate simulated movement from initial positions and conditions.

        This is an abstract method that must be implemented in any subclass.

        Arguments
        ---------
        initial_position : array
            Array of initial positions of shape [num, 2] to specify coordinates
            of individuals to be simulated.
        site: :py:obj:`ollin.Site`
            Site in which to simulate movement.
        steps : int
            Number of steps to simulate.
        velocity : int
            Mean velocity of individuals.

        Returns
        -------
        array : array
            Array of shape [num, steps, 2], so if (x, y) = Array[i, j, :] then
            x and y are the coordinates of the i-th individual at step j in the
            simulation. Number of steps is determined by number of days and
            steps per day parameter.
        """
        pass



from __future__ import print_function
from __future__ import division
from multiprocessing import Pool

import sys
import numpy as np
import ollin

from ..core.utils import density_to_occupancy, logit
from ..core.constants import GLOBAL_CONSTANTS


TRIALS_PER_WORLD = 100
MAX_INDIVIDUALS = 10000
NUM_WORLDS = 10
HOME_RANGES = np.linspace(0.1, 3, 6)
NICHE_SIZES = np.linspace(0.3, 0.9, 6)
NUMS = np.linspace(10, 1000, 6, dtype=np.int64)


class OccupancyCalibrator(object):
    def __init__(
            self,
            movement_model,
            home_ranges=HOME_RANGES,
            niche_sizes=NICHE_SIZES,
            nums=NUMS,
            trials_per_world=TRIALS_PER_WORLD,
            num_worlds=NUM_WORLDS,
            range=GLOBAL_CONSTANTS['range'],
            season=GLOBAL_CONSTANTS['season'],
            max_individuals=MAX_INDIVIDUALS):

        self.movement_model = movement_model
        self.home_ranges = home_ranges
        self.niche_sizes = niche_sizes
        self.nums = nums
        self.trials_per_world = trials_per_world
        self.num_worlds = num_worlds
        self.max_individuals = max_individuals
        self.season = season
        if isinstance(range, (int, float)):
            range = (range, range)
        self.range = range

        self.oc_info = self.calculate_oc_info()

    def calculate_oc_info(self):
        n_hr = len(self.home_ranges)
        n_nsz = len(self.niche_sizes)
        n_num = len(self.nums)
        tpw = self.trials_per_world
        nw = self.num_worlds
        mov = self.movement_model
        mx_ind = self.max_individuals
        season = self.season

        all_info = np.zeros(
                [n_hr, n_nsz, n_num, nw, tpw])
        arguments = [
                Info(mov, hr, nsz, self.nums, tpw, self.range, season, mx_ind)
                for hr in self.home_ranges
                for nsz in self.niche_sizes
                for k in xrange(self.num_worlds)]

        nargs = len(arguments)
        msg = 'Making {} runs of the simulator'
        msg += '\n\tSimulating a total of {} individuals'
        msg = msg.format(nargs, n_hr * n_nsz * tpw * np.sum(self.nums))
        print(msg)
        pool = Pool()
        try:
            results = pool.map(get_single_oc_info, arguments)
            pool.close()
            pool.join()
        except KeyboardInterrupt:
            pool.terminate()
            sys.exit()
            quit()

        arguments = [
                (i, j, k)
                for i in xrange(n_hr)
                for j in xrange(n_nsz)
                for k in xrange(self.num_worlds)]

        for (i, j, k), res in zip(arguments, results):
            all_info[i, j, :, k, :] = res

        return all_info

    def plot(
            self,
            figsize=(10, 10),
            ax=None,
            x_var='density',
            w_target=True,
            xscale=None,
            yscale=None,
            beta=1,
            lwidth=0.1,
            wtext=False,
            wspace=0):
        import matplotlib.pyplot as plt
        from matplotlib.ticker import NullFormatter

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        area = self.range[0] * self.range[1]

        density = self.nums / area
        hr_proportions = self.home_ranges / area

        if x_var == 'density':
            iterator1 = hr_proportions
            var2 = 'HRP'
            iterator2 = self.niche_sizes
            var3 = 'NS'
        elif x_var == 'home_range':
            iterator1 = density
            var2 = 'D'
            iterator2 = self.niche_sizes
            var3 = 'NS'
        elif x_var == 'niche_sizes':
            iterator1 = hr_proportions
            var2 = 'HRP'
            iterator2 = density
            var3 = 'D'

        ncols = len(iterator1)
        nrows = len(iterator2)
        params = self.movement_model.parameters['density']

        counter = 1
        for m, x in enumerate(iterator1):
            for n, y in enumerate(iterator2):
                nax = plt.subplot(nrows, ncols, counter)

                if x_var == 'density':
                    data = self.oc_info[m, n, :, :, :]
                    xcoords = density
                elif x_var == 'home_range':
                    data = self.oc_info[:, n, m, :, :]
                    xcoords = beta * hr_proportions
                elif x_var == 'niche_sizes':
                    data = self.oc_info[m, :, n, :, :]
                    xcoords = self.niche_sizes

                mean = data.mean(axis=(1, 2))
                std = data.std(axis=(1, 2))
                uplim = mean + std
                dnlim = mean - std

                xtext = 0.1
                ytext = 0.8

                ylim0, ylim1 = -0.1, 1.1

                if xscale == 'log':
                    xcoords = np.log(xcoords)
                    xtext = np.log(xtext)

                if yscale == 'log':
                    mean = np.log(mean)
                    uplim = np.log(uplim)
                    dnlim = np.log(dnlim)
                    ylim0 = -6
                    ylim1 = 0
                    ytext = np.log(ytext)

                if yscale == 'logit':
                    mean = logit(mean)
                    uplim = logit(uplim)
                    dnlim = logit(dnlim)
                    ylim0 = -6
                    ylim1 = 4
                    ytext = logit(ytext)

                nax.plot(
                    xcoords,
                    mean,
                    linewidth=lwidth)
                nax.fill_between(
                    xcoords,
                    dnlim,
                    uplim,
                    alpha=0.6,
                    edgecolor='white')

                if w_target:
                    if x_var == 'density':
                        target = density_to_occupancy(
                            density,
                            x,
                            y,
                            parameters=params)
                    elif x_var == 'home_range':
                        target = density_to_occupancy(
                            x,
                            hr_proportions,
                            y,
                            parameters=params)
                    elif x_var == 'niche_sizes':
                        target = density_to_occupancy(
                            y,
                            x,
                            self.niche_sizes,
                            parameters=params)

                    if yscale == 'log':
                        target = np.log(target)
                    if yscale == 'logit':
                        target = logit(target)

                    nax.plot(
                        xcoords,
                        target,
                        color='red',
                        label='target')

                nax.set_ylim(ylim0, ylim1)
                nax.set_xlim(xcoords.min(), xcoords.max())

                if wtext:
                    nax.text(
                        xtext, ytext, '{}={}\n{}={}'.format(var2, x, var3, y))

                if m == ncols - 1:
                    nax.set_xlabel('{}={}'.format(var3, y))
                if n == 0:
                    nax.set_ylabel('{}={}'.format(var2, x))

                if m < ncols - 1:
                    nax.xaxis.set_major_formatter(NullFormatter())
                if n > 0:
                    nax.yaxis.set_major_formatter(NullFormatter())

                counter += 1
        plt.subplots_adjust(wspace=wspace, hspace=wspace)

        font = {'fontsize': 18}
        plt.figtext(0.4, 0.035, x_var, fontdict=font)
        plt.figtext(0.035, 0.5, "Occupancy (%)", fontdict=font, rotation=90)
        title = "Occupancy Calibration\n{}"
        title = title.format(self.movement_model.name)
        plt.figtext(0.38, 0.92, title, fontdict=font)
        return ax

    def fit(self):
        from sklearn.linear_model import LinearRegression
        data = self.oc_info
        area = self.range[0] * self.range[1]
        density = self.nums / float(area)
        hr_proportions = self.home_ranges / float(area)

        X = []
        Y = []
        for i, nsz in enumerate(self.niche_sizes):
            for j, hr in enumerate(hr_proportions):
                for k, dens in enumerate(density):
                    oc_data = data[j, i, k, :, :].ravel()
                    hr_data = hr * np.ones_like(oc_data)
                    dens_data = dens * np.ones_like(oc_data)
                    nsz_data = nsz * np.ones_like(oc_data)
                    Y.append(logit(oc_data))
                    X.append(
                        np.stack([np.log(hr_data),
                                  np.log(dens_data),
                                  np.log(nsz_data)], -1))
        X = np.concatenate(X, 0)
        Y = np.concatenate(Y, 0)

        lrm = LinearRegression()
        lrm.fit(X, Y)

        alpha = lrm.intercept_
        hr_exp = lrm.coef_[0]
        den_exp = lrm.coef_[1]
        nsz_exp = lrm.coef_[2]

        parameters = {
            'alpha': alpha,
            'hr_exp': hr_exp,
            'density_exp': den_exp,
            'niche_size_exp': nsz_exp}
        return parameters


class Info(object):
    __slots__ = [
        'movement_model',
        'home_range',
        'niche_size',
        'nums',
        'trials',
        'range',
        'season',
        'max_individuals']

    def __init__(
            self,
            movement_model,
            home_range,
            niche_size,
            nums,
            trials,
            range_,
            season,
            max_individuals):

        self.movement_model = movement_model
        self.home_range = home_range
        self.niche_size = niche_size
        self.nums = nums
        self.max_individuals = max_individuals
        self.season = season
        self.trials = trials
        self.range = range_


def get_single_oc_info(info):
    site = ollin.Site.make_random(info.niche_size, range=info.range)
    mov = ollin.Movement.simulate(
        site,
        num=info.max_individuals,
        home_range=info.home_range,
        days=info.season,
        movement_model=info.movement_model)

    n_nums = len(info.nums)
    results = np.zeros([n_nums, info.trials])

    for n, num in enumerate(info.nums):
        for k in xrange(info.trials):
            submov = mov.sample(num)
            oc = ollin.Occupancy(submov)
            results[n, k] = oc.occupancy

    return results



from six.moves import xrange
import numpy as np
import math
from numba import jit, float64, int64

from .basemodel import MovementModel


class Model(MovementModel):
    name = 'Constant Levy Model'
    default_parameters = {
        'movement': {
            'pareto': 1.8},
        "density": {
            "alpha": 5.139654979615388,
            "density_exp": 0.8673231302202313,
            "hr_exp": 0.8709250205414867,
            "niche_size_exp": 0.8105129468063047
            },
        "home_range": {
            "alpha": 82.0335925804581,
            "exponent": 1.9018678531419373
            },
        "velocity": {
            "alpha": 0.0005826019701355142,
            "beta": 1.0525954033363216
            }
    }

    def generate_movement(
            self,
            initial_positions,
            site,
            steps,
            velocity):
        exponent = self.parameters['movement']['pareto']
        range_ = site.range

        mov = self._movement(
            initial_positions,
            velocity,
            range_,
            steps,
            exponent)
        return mov

    @staticmethod
    @jit(
        float64[:, :, :](
            float64[:, :],
            float64,
            float64[:],
            int64,
            float64),
        nopython=True)
    def _movement(
            random_positions,
            velocity,
            range_,
            steps,
            exponent):
        num, _ = random_positions.shape
        movement = np.zeros((num, steps, 2), dtype=float64)
        random_angles = np.random.uniform(0.0, 2 * np.pi, size=(steps, num))
        rangex, rangey = range_
        for k in xrange(steps):
            movement[:, k, :] = random_positions
            for j in xrange(num):
                angle = random_angles[k, j]
                heading = (math.cos(angle), math.sin(angle))
                magnitude = (velocity * (exponent - 1)) / \
                    (math.pow((1 - np.random.rand()), 1/exponent) * exponent)
                direction = (magnitude * heading[0], magnitude * heading[1])
                tmp1 = (
                    random_positions[j, 0] + direction[0],
                    random_positions[j, 1] + direction[1])
                tmp2 = (tmp1[0] % (2 * rangex), tmp1[1] % (2 * rangey))
                if tmp2[0] < rangex:
                    random_positions[j, 0] = tmp2[0] % rangex
                else:
                    random_positions[j, 0] = (-tmp2[0]) % rangex
                if tmp2[1] < rangey:
                    random_positions[j, 1] = tmp2[1] % rangey
                else:
                    random_positions[j, 1] = (-tmp2[1]) % rangey
        return movement



"""Module defining Movement Data class and Movement Analyzer

Either simulated data or data incoming from real telemetry data can be stored
in a :py:class:`MovementData` object. The main information held in such an
object is the full history of individual positions arranged in an array of
shape [num_individuals, time_steps, 2]. This information can then be
plotted for trajectory visualization, or used in further processing.

Data produced by simulation can be stored in a specialized type of MovementData
which also holds movement model information. See :py:class:`Movement`.

Movement analysis, such as distribution of velocities, heading angles and
turning angles can be extracted and stored in an :py:class:`MovementAnalysis`
object.
"""
from __future__ import division
from __future__ import print_function
import copy

import numpy as np

from ..movement_models.basemodel import MovementModel
from .constants import GLOBAL_CONSTANTS
from ..movement_models import get_movement_model
from .utils import (occupancy_to_density,
                    home_range_to_velocity,
                    velocity_modification)


class MovementData(object):
    """Container for Movement data.

    All animal movement data can be stored in an array of shape of shape
    [num_individuals, time_steps, 2] which represents the positions of every
    individual along some time interval. If::

        x = array[i, j, 0]
        y = array[i, j, 1]

    then the i-th individual was at the place with (x, y)-coordinates at the
    j-th time step.

    Apart from spatial information, times at which the timesteps where taken
    are store in another array of shape [time_steps].

    Attributes
    ----------
    site : :py:obj:`ollin.Site`
        Information of Site at which movement took place.
    movement_data : array
        Array of shape [num_individuals, time_steps, 2] holding coordinate
        information of individual location through movement.
    times : array
        Array of shape [time_steps] with time at which the time steps took
        place. Units are in days.
    home_range : float or None
        Home range value of species. Only necessary for occupancy calculation.
        See :py:class:`ollin.Occupancy`.

    """
    def __init__(self, site, movement_data, times, home_range=None):
        """Construct Movement Data object.

        Arguments
        ---------
        site : :py:obj:`ollin.Site`
            Information of Site at which movement took place.
        movement_data : array
            Array of shape [num_individuals, time_steps, 2] holding coordinate
            information of individual location through movement.
        times : array
            Array of shape [time_steps] with time at which the time steps took
            place. Units are in days.
        home_range : float, optional
            Home range value of species. Only necessary for occupancy
            calculation. See :py:class:`ollin.Occupancy`.

        """
        self.site = site
        self.data = movement_data
        self.times = times
        self.home_range = home_range
        self.num, self.steps, _ = movement_data.shape

    def num_slice(self, key):
        """Extract motion from slice of individuals.

        Select a subset of individuals from motion data using a
        slice.

        Arguments
        ---------
            key : int or list or tuple or slice
                If key is an integer the result will be a
                :py:obj:`MovementData` object holding only motion data for the
                corresponding individual. If key is a list or tuple, its
                contents will be passed to the :py:func:`slice` function, and
                said slice will be extracted from data array in the first axis,
                and returned in an :py:obj:`MovementData` object.

        Returns
        -------
            newcopy : :py:obj:`MovementData`
                New :py:obj:`MovementData` object sharing site and times
                attributes but with movement data corresponding to individuals
                slice.

        Example
        -------
        To extract the movement of the first ten individuals::

            first_ten = movement.num_slice((None, 10, None))

        To extract the movement of the last 20 individuals::

            last_20 = movement.num_slice((-20, None, None))

        To extract all even individuals::

            even = movement.num_slice((None, None, 2))

        """
        if not isinstance(key, (int, slice)):
            if isinstance(key, (list, tuple)):
                key = slice(*key)
            else:
                msg = 'Num slice only accepts (int/list/tuple/slice) as'
                msg += ' arguments. {} given.'.format(type(key))
                raise ValueError(msg)
        data = self.data[key, :, :]

        newcopy = copy.copy(self)
        newcopy.data = data
        newcopy.num, newcopy.steps, _ = data.shape
        return newcopy

    def sample(self, num):
        """Extract a sample of individual movement.

        Select a random sample of individuals of a given size to form a new
        :py:obj:`MovementData` object.

        Arguments
        ---------
        num : int
            Size of sample

        Returns
        -------
        newcopy : :py:obj:`MovementData`
            Movement data corresponding to sample.

        """
        selection = np.random.choice(
            np.arange(self.num),
            size=num)
        data = self.data[selection, :, :]
        newcopy = copy.copy(self)
        newcopy.data = data
        newcopy.num, newcopy.steps, _ = data.shape
        return newcopy

    def select(self, selection):
        """Select a subset of individual movement.

        Use an array of indices to select a subset of individuals and return
        movement data of the corresponding individuals.

        Arguments
        ---------
        selection : array or tuple or list
            List of indices of selected individuals

        Returns
        -------
        newcopy : :py:obj:`MovementData`
            Movement data of selected individuals.

        """
        if isinstance(selection, (tuple, list)):
            selection = np.array(selection)
        data = self.data[selection, :, :]
        newcopy = copy.copy(self)
        newcopy.data = data
        newcopy.num, newcopy.steps, _ = data.shape
        return newcopy

    def time_slice(self, key):
        """Select a slice of timesteps from movement.

        Arguments
        ---------
        key : int or list or tuple or slice
            If key is integer the resulting :py:obj:`MovementData` object will
            only hold the individuals position at the corresponding timestep.
            If key is list or tuple, its contents will be passed to the
            :py:func:slice function and the slice will be used to extract some
            times steps from the movement data.

        Returns
        -------
        newcopy : :py:obj:`MovementData`
            Movement data with selected time steps.

        Example
        -------
        To select the first 10 days of movement::

            first_10_days = movement_data.time_slice((None, 10, None))

        To select the last 20 days of movement::

            last_20_days = movement_data.time_slice((-20, None, None))

        To select every other step::

            every_other = movement_data.time_slice((None, None, 2))

        """
        if not isinstance(key, (int, slice)):
            if isinstance(key, (list, tuple)):
                key = slice(*key)
            else:
                msg = 'Time slice only accepts (int/list/tuple/slice) as'
                msg += ' arguments. {} given.'.format(type(key))
                raise ValueError(msg)

        data = self.data[:, key, :]
        newcopy = copy.copy(self)
        newcopy.data = data
        newcopy.num, newcopy.steps, _ = data.shape
        return newcopy

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            num=10,
            steps=1000,
            mov_cmap='Greens',
            simplify=None,
            **kwargs):
        """Plot trajectories from Movement data.

        Movement Data plotting adds the following optional components to the
        plot:

        1. "trajectories":
           If present in include list, some trajectories will be plotted as a
           broken line. Trajectory simplification is possible through the
           simplify keyworded argument. Several trajectories will be plotted.
           Color of line will be chosen at random from some colormap.

        All other components in the include list will be passed down to the
        Site plotting method. See :py:meth:`ollin.Site.plot` for all plot
        components defined at that level.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot movement trajectories.
        figsize : list or tuple, optional
            Size of figure to create if no axes object was given. Defaults to
            (10, 10).
        include : list or tuple, optional
            List of components to add to the plot. Components list will be
            passed to the Site object to add the corresponding components.
        num : int, optional
            Number of trajectories to plot. Defaults to 10.
        steps : int, optional
            Number of time steps to plot in trajectories. Defaults to all.
        mov_cmap : str, optional
            Name of colormap to choose trajectories colors from. See
            :py:mod:`matplotlib.cm` to see all options. Defaults to 'Greens'.
        simplify : int, optional
            Trajectories will be forced to consist of this number of points, so
            if given, some time steps might be skipped.
        kwargs : dict, optional
            All other keyworded arguments will be passed to the Site's plotting
            method.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Return axes for further plotting.

        """
        import matplotlib.pyplot as plt  # pylint: disable=import-error
        from cycler import cycler

        if include is None:
            include = [
                'niche',
                'niche_boundary',
                'rectangle',
                'trajectories']

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        self.site.plot(
            include=include, ax=ax, **kwargs)

        if 'trajectories' in include:
            cmap = plt.get_cmap(mov_cmap)
            colors = [cmap(i) for i in np.linspace(.8, .1, 10)]
            ax.set_prop_cycle(cycler('color', colors))

            steps = min(self.steps, steps)

            if simplify is None:
                stride = 1
            else:
                stride = max(int(steps / simplify), 1)
            trajectories = self.data[:num, :steps:stride, :]

            for trajectory in trajectories:
                xcoord, ycoord = zip(*trajectory)
                ax.plot(xcoord, ycoord)

        return ax

    def analyze(self):
        """Analyze movement and return analysis."""
        return MovementAnalysis(self)


class Movement(MovementData):
    """Class for simulated movement data.

    Extension of :py:class:`MovementData` class. When movement data arises from
    simulation, the applied movement model is also stored within the object.

    Attributes
    ----------
    site : :py:obj:`ollin.Site`
        Information of Site at which movement took place.
    movement_data : array
        Array of shape [num_individuals, time_steps, 2] holding coordinate
        information of individual location through movement.
    times : array
        Array of shape [time_steps] with time at which the time steps took
        place. Units are in days.
    home_range : float or None
        Home range value of species. Only necessary for occupancy calculation.
        See :py:class:`ollin.Occupancy`.
    movement_model : :py:obj:`ollin.MovementModel`
        Movement model used to generate movement.
    velocity : float
        Mean velocity (in Km/Day) used to movement simulation.

    """

    def __init__(
            self,
            site,
            movement_data,
            movement_model,
            velocity,
            home_range=None):
        """Create Movement object for simulated movement.

        Arguments
        ---------
        site : :py:obj:`ollin.Site`
            Site in which movement took place.
        movement_data : array
            Array of shape [num_individuals, time_steps, 2] holding coordinate
            information of all individuals along all simulated time steps.
        movement_model : :py:obj:`ollin.MovementModel`
            Movement model used to generate movement_data.
        velocity : float
            Mean velocity (in Km/Day) used in the simulation.
        home_range : float, optional
            Home range of simulated species. Used mainly for occupancy
            calculation, or home range calibration. See
            :py:class:`ollin.Occupancy`.

        """
        self.movement_model = movement_model
        self.velocity = velocity

        steps = movement_data.shape[1]
        steps_per_day = movement_model.parameters['steps_per_day']
        days = steps / steps_per_day
        times = np.linspace(0, days, steps)

        super(Movement, self).__init__(
            site, movement_data, times, home_range=home_range)

    @classmethod
    def simulate(
            cls,
            site,
            days=None,
            num=None,
            occupancy=None,
            home_range=None,
            velocity=None,
            parameters=None,
            movement_model='variable_levy'):
        """Make simulated movement data.

        Use some movement model from the model library to generate simulated
        movement data for some virtual species with a fixed velocity (in
        Km/Day).

        Number of individuals and mean velocity must be specified, but it is
        also possible to use home range and/or occupancy as proxies for density
        and mean velocity, respectively. The faithfullness of these proxies
        depend on the correct calibration of the parameters asociated to the
        movement model.

        If using a movement model from the library, these should be
        pre-calibrated, and hence home range (or occupancy) can be used to
        estimate mean velocity (or density) with some degree of accuracy.

        Otherwise the user must first be sure that the model is calibrated.
        See :py:mod:`ollin.calibration`.

        Arguments
        ---------
        site : :py:obj:`ollin.Site`
            Site in which simulate movement.
        days : int, optional
            Number of simulation days. Movement models include a steps_per_day
            parameter, so number of simulated time steps is days *
            steps_per_day. Defaults to 365.
        num : int, optional
            Number of individuals to include in simulation. If not given,
            occupancy argument must be provided.
        occupancy : float, optional
            If provided the relationship occupancy <-> density will be used to
            estimate the number of individuals to include in simulation. See
            :py:func:`ollin.core.utils.occupancy_to_density`.
        velocity : float, optional
            Mean velocity in Km/Day to use in movement model. If not given,
            home range argument must be provided.
        home_range : float, optional
            Home range of simulated species. If provided the relationship
            home_range <-> mean velocity will be used to estimate the mean
            velocity of species. See
            :py:func:`ollin.core.utils.home_range_to_velocity`.
        movement_model : str or :py:obj:`ollin.movement_models.MovementModel`
            Name of movement model in library o MovementModel instance to use
            to generate simulated movement.

        Returns
        -------
        mov : :py:obj:`Movement`
            Movement instance with simulated movement data.

        Raises
        ------
        ValueError
            If both num and occupancy, or velocity and home_range, are given
            simultaneously.

        """
        if not isinstance(movement_model, MovementModel):
            movement_model = get_movement_model(
                movement_model,
                parameters=parameters)
        parameters = movement_model.parameters

        if velocity is None:
            if home_range is None:
                msg = 'Arguments velocity or home_range must be provided'
                raise ValueError(msg)
            velocity = home_range_to_velocity(
                home_range,
                parameters=parameters['home_range'])

        if num is None:
            if occupancy is None:
                msg = 'Arguments num or occupancy must be provided'
                raise ValueError(msg)
            rangex, rangey = site.range
            if home_range is None:
                msg = 'If num is not specified home range AND occupancy'
                msg += ' must be provided'
                raise ValueError(msg)
            area = site.range[0] * site.range[1]
            home_range_proportion = home_range / area
            dens = occupancy_to_density(
                occupancy,
                home_range_proportion,
                site.niche_size,
                parameters=parameters['density'])
            num = int(rangex * rangey * dens)

        if days is None:
            days = GLOBAL_CONSTANTS['days']

        velocity_mod = velocity_modification(
            site.niche_size, parameters)
        steps_per_day = parameters['steps_per_day']
        sim_velocity = velocity * velocity_mod / steps_per_day

        steps = int(days * steps_per_day)

        initial_positions = site.sample(num)
        movement_data = movement_model.generate_movement(
            initial_positions,
            site,
            steps,
            sim_velocity)

        return cls(
            site,
            movement_data,
            movement_model,
            velocity,
            home_range=home_range)

    def extend(self, days, inplace=True):
        """Extend movement data with new simulated movement.

        Use last position as starting point to generate new simulated
        movement and append to existing. This method will use the same mean
        velicity and movement model to generate new movement.

        Arguments
        ---------
        days : int
            Number of days of new simulated movement.
        inplace : bool, optional
            If true, only Movement object attributes will be changed, otherwise
            a copy of the object will be made with the new movement data.

        Returns
        -------
        extension : :py:obj:`Movement`
            Movement object with extended movement data.

        """

        parameters = self.movement_model.parameters
        steps_per_day = parameters['steps_per_day']
        steps = int(steps_per_day * days)

        velocity_mod = velocity_modification(
            self.site.niche_size, parameters)
        velocity = self.velocity * velocity_mod / steps_per_day

        initial_positions = self.data[:, -1, :]

        new_data = self.movement_model.generate_movement(
            initial_positions,
            self.site,
            steps + 1,
            velocity)
        data = np.append(
            self.data, new_data[:, 1:, :], 1)

        old_steps = self.data.shape[1]
        total_days = (old_steps + steps) / steps_per_day
        times = np.linspace(0, total_days, old_steps + steps)

        if inplace:
            extension = self
        else:
            extension = copy.copy(self)

        extension.data = data
        extension.times = times
        return extension


class MovementAnalysis(object):
    """Class for movement analysis.

    Movement analysis refers to the calculation of the following information:

    1. Velocities:
        Information of time at which time steps where taken with distance
        travelled can be used to calculate velocity of every individual at
        each timestep.

    2. Bearings:
        Bearing refers to the angle of direction at some time step. The angle
        is taken in reference to the x-axis. The units are radians and the
        angle value ranges from [-pi, pi]. Bearing values can be calculated for
        every individual at each timestep.

    3. Turn Angles:
        Turn angle refers to angle formed between two adjacent bearing
        directions in the same trajectory. The turn angle is in radians and its
        values ranges from [-pi, pi]. Turn angle values can be calculated for
        every individual at each timestep.

    MovementAnalysis objects hold this information.

    Attributes
    ----------
    movement : :py:obj:`ollin.MovementData`
        Movement data analyzed.
    velocities : array
        Array of shape [num_individuals, time_steps - 1] holding all velocity
        information.::

            velocities[i, j] = v

        Means that the i-th individual had velocity v at the j-th time step.
    bearings : array
        Array of shape [num_individuals, time_steps - 1] holding all bearing
        information.
    turn_angles : array
        Array of shape [num_individuals, time_steps - 2] holding all turn angle
        information.

    """

    def __init__(self, mov):
        """Analyze movement object.

        Arguments
        ---------
        mov : :py:obj:`MovementData`
            Movement data to analyze.
        """

        self.movement = mov
        self.velocities, self.bearings, self.turn_angles = self.analyze(mov)

    def get_mean_velocity(self):
        """Calculate and return mean velocity"""
        return self.velocities.mean()

    @staticmethod
    def analyze(movement):
        """Calculate and return velocities, bearings and turn angles"""
        data = movement.data
        times = movement.times

        dtimes = (times[1:] - times[:-1])[None, :, None]
        directions = (data[:, 1:, :] - data[:, :-1, :]) / dtimes
        complex_directions = directions[:, :, 0] + 1j * directions[:, :, 1]
        velocities = np.abs(complex_directions)
        bearings = np.angle(complex_directions)
        turn_angles = np.angle(complex_directions[:, 1:] /
                               complex_directions[:, :-1])
        return velocities, bearings, turn_angles

    def plot_velocity_distribution(
            self,
            ax=None,
            figsize=(10, 10),
            num_individual=0,
            bins=20,
            width=None,
            cmap='Reds',
            alpha=0.8,
            log=True):
        """Plot distribution of velocities.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot.
        figsize : tuple or list, optional
            Size of figure to create if no axes are provided.
        num_individual : int or tuple or list or array, optional
            Selection of individuals from which to draw velocity
            information. If num_individual='all', all information will
            be plotted.
        bins : int, optional
            Number of bins to use in histogram of velocity distribution.
            Defaults to 20.
        width : float, optional
            Width of bars in histogram. If none is given, width will be
            maximum possible before overlap.
        cmap : str, optional
            Colormap to use to assign colors to histogram bars. Defaults
            to 'Reds'.
        alpha : float, optional
            Alpha value of plot. Defaults to 0.8.
        log : bool, optional
            If true, yaxis in histogram will have logarithmic scale.
        """
        import matplotlib.pyplot as plt
        from matplotlib.cm import get_cmap

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        if num_individual == 'all':
            vdata = self.velocities.ravel()
        elif isinstance(num_individual, int):
            vdata = self.velocities[num_individual, :]
        else:
            vdata = self.velocities[num_individual, :].ravel()

        range_ = (0, vdata.max())
        histogram, _ = np.histogram(
            vdata, bins=bins, range=range_, normed=not log)
        if log:
            histogram = np.log(histogram + 1)

        if width is None:
            width = (range_[1] - range_[0]) / bins
        theta = np.linspace(range_[0], range_[1], bins, endpoint=False)

        bars = ax.bar(theta, histogram, width=width, bottom=0)
        mins, maxs = histogram.min(), histogram.max()

        # Use custom colors and opacity
        cmap = get_cmap(cmap)
        for value, pbar in zip(histogram, bars):
            pbar.set_facecolor(
                cmap((value - mins + 0.1) / (0.1 + maxs - mins)))
            pbar.set_alpha(alpha)
        ticks = np.linspace(0, histogram.max(), 10)
        ax.set_yticks(ticks)
        ax.set_yticklabels(np.round(ticks, 2))

        ax.set_title('Velocity distribution')
        ax.set_xlabel('Velocity (Km/Day)')

        if log:
            ax.set_ylabel('Log count')
        else:
            ax.set_ylabel('Proportion')

        return ax

    def plot_bearing_distribution(
            self,
            ax=None,
            figsize=(10, 10),
            num_individual=0,
            bins=20,
            width=None,
            cmap='Reds',
            alpha=0.8):
        """Plot distribution of bearing angles.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot.
        figsize : tuple or list, optional
            Size of figure to create if no axes are provided.
        num_individual : int or tuple or list or array, optional
            Selection of individuals from which to draw bearing angle
            information. If num_individual='all', all information will
            be plotted.
        bins : int, optional
            Number of bins to use in histogram of bearing distribution.
            Defaults to 20.
        width : float, optional
            Width of bars in histogram. If none is given, width will be
            maximum possible before overlap.
        cmap : str, optional
            Colormap to use to assign colors to histogram bars. Defaults
            to 'Reds'.
        alpha : float, optional
            Alpha value of plot. Defaults to 0.8.
        """
        import matplotlib.pyplot as plt
        from matplotlib.cm import get_cmap

        if ax is None:
            _, ax = plt.subplots(
                figsize=figsize, subplot_kw={"polar": True})

        if num_individual == 'all':
            bdata = self.bearings.ravel()
        elif isinstance(num_individual, int):
            bdata = self.bearings[num_individual, :]
        else:
            bdata = self.bearings[num_individual, :].ravel()

        range_ = (-np.pi, np.pi)
        histogram, _ = np.histogram(
            bdata, bins=bins, range=range_, normed=True)

        if width is None:
            width = (range_[1] - range_[0]) / bins
        theta = np.linspace(range_[0], range_[1], bins, endpoint=False)

        bars = ax.bar(theta, histogram, width=width, bottom=0.05)
        mins, maxs = histogram.min(), histogram.max()

        # Use custom colors and opacity
        cmap = get_cmap(cmap)
        for value, pbar in zip(histogram, bars):
            pbar.set_facecolor(
                cmap((value - mins + 0.1) / (0.1 + maxs - mins)))
            pbar.set_alpha(alpha)
        ticks = np.linspace(0.05, 0.05 + histogram.max(), 4)
        ax.set_yticks(ticks)
        ax.set_yticklabels(np.round(ticks - 0.05, 2))

        ax.set_title('Bearing distribution')
        return ax

    def plot_turn_angle_distribution(
            self,
            ax=None,
            figsize=(10, 10),
            num_individual=0,
            bins=20,
            width=None,
            cmap='Reds',
            alpha=0.8):
        """Plot distribution of turning angles.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot.
        figsize : tuple or list, optional
            Size of figure to create if no axes are provided.
        num_individual : int or tuple or list or array, optional
            Selection of individuals from which to draw turning angle
            information. If num_individual='all', all information will
            be plotted.
        bins : int, optional
            Number of bins to use in histogram of turn angle distribution.
            Defaults to 20.
        width : float, optional
            Width of bars in histogram. If none is given, width will be
            maximum possible before overlap.
        cmap : str, optional
            Colormap to use to assign colors to histogram bars. Defaults
            to 'Reds'.
        alpha : float, optional
            Alpha value of plot. Defaults to 0.8.
        """
        import matplotlib.pyplot as plt
        from matplotlib.cm import get_cmap

        if ax is None:
            _, ax = plt.subplots(
                figsize=figsize, subplot_kw={"polar": True})

        if num_individual == 'all':
            tdata = self.turn_angles.ravel()
        elif isinstance(num_individual, int):
            tdata = self.turn_angles[num_individual, :]
        else:
            tdata = self.turn_angles[num_individual, :].ravel()

        range_ = (-np.pi, np.pi)
        histogram, _ = np.histogram(
            tdata, bins=bins, range=range_, normed=True)

        if width is None:
            width = (range_[1] - range_[0]) / bins
        theta = np.linspace(range_[0], range_[1], bins, endpoint=False)

        bars = ax.bar(theta, histogram, width=width, bottom=0.05)
        mins, maxs = histogram.min(), histogram.max()

        # Use custom colors and opacity
        cmap = get_cmap(cmap)
        for value, pbar in zip(histogram, bars):
            pbar.set_facecolor(
                cmap((value - mins + 0.1) / (0.1 + maxs - mins)))
            pbar.set_alpha(alpha)
        ticks = np.linspace(0.05, 0.05 + histogram.max(), 4)
        ax.set_yticks(ticks)
        ax.set_yticklabels(np.round(ticks - 0.05, 2))

        ax.set_title('Turn Angle distribution')
        return ax



from ..stanmodels import StanModel
from .occupancy_estimation import OccupancyEstimate

CODE = u'''
data {
    int<lower=0> steps;
    int<lower=0> cams;
    int<lower=0, upper=1> detections[cams, steps];

    // Parameters for Ocuppancy Beta Prior
    real<lower=0> alpha_oc;
    real<lower=0> beta_oc;

    // Parameters for detectionsectability Beta Prior
    real<lower=0> alpha_det;
    real<lower=0> beta_det;
}
transformed data {
    int<lower=0> counts[cams]; // detectionsection count per site
    for(m in 1:cams) {
        counts[m] = sum(detections[m]);
    }
}
parameters {
    real<lower=0, upper=1> occupancy;
    real<lower=0, upper=1> detectability;
}
model {
    occupancy ~ beta(alpha_oc, beta_oc);
    detectability ~ beta(alpha_det, beta_det);

    for (j in 1:cams) {
        if (counts[j] == 0)
            target += log_sum_exp(
                binomial_lpmf(counts[j] | steps, detectability) + log(occupancy),
                log(1 - occupancy));
        else
            target += log(occupancy) + binomial_lpmf(counts[j] | steps, detectability);
    }
}
'''


class Model(StanModel):
    name = 'McKenzie Single Species - Single Season'
    stancode = CODE

    def prepare_data(self, detection, priors):
        steps, cams = detection.detections.shape
        data = {
            'steps': steps,
            'cams': cams,
            'detections': detection.detections.T.astype(int).tolist(),
            'alpha_det': priors.get('alpha_det', 1),
            'beta_det': priors.get('beta_det', 1),
            'alpha_oc': priors.get('alpha_oc', 1),
            'beta_oc': priors.get('beta_oc', 1)}
        return data

    def estimate(self, detection, method='MAP', priors=None):
        stan_result = super(Model, self).estimate(detection, method, priors)
        occupancy = stan_result['occupancy']
        detectability = stan_result['detectability']
        est = OccupancyEstimate(
            occupancy, self, detection, detectability=detectability)
        return est



from __future__ import print_function

from abc import abstractmethod
import os
import pickle

from .estimation import EstimationModel


COMPILED_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'compiled_models')


class StanModel(EstimationModel):
    name = None
    stancode = None

    def __init__(self):
        self.stanmodel = self.load_model()

    @abstractmethod
    def prepare_data(self, detection, priors):
        pass

    @abstractmethod
    def estimate(self, detection, method='MAP', priors=None):
        if priors is None:
            priors = {}

        model = self.stanmodel
        data = self.prepare_data(detection, priors)

        if method == 'MAP':
            op = model.optimizing(data=data)
        elif method == 'mean':
            op = model.sampling(data=data)
        else:
            msg = 'Estimation method {} not implemented.'.format(method)
            raise NotImplementedError(msg)
        return op

    def load_model(self):

        if not os.path.exists(COMPILED_PATH):
            os.makedirs(COMPILED_PATH)

        path = os.path.join(
            COMPILED_PATH, self.name.replace(' ', '_') + '.pkl')
        if os.path.exists(path):
            with open(path, 'rb') as stanfile:
                stan_model = pickle.load(stanfile)
        else:
            stan_model = self.compile_and_save()
        return stan_model

    def compile_and_save(self):
        import pystan

        stan_model = pystan.StanModel(model_code=self.stancode)

        path = os.path.join(
            COMPILED_PATH, self.name.replace(' ', '_') + '.pkl')
        with open(path, 'wb') as stanfile:
            pickle.dump(stan_model, stanfile)

        return stan_model



"""
"""
from abc import abstractmethod
from importlib import import_module
import os
import glob

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


class Estimate(object):
    def __init__(self, model, data):
        self.model = model
        self.data = data


class EstimationModel(object):
    name = None

    @abstractmethod
    def estimate(self, detection):
        pass


@lru_cache()
def get_estimation_model(what, name):
    path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        what,
        name + '.py')
    if os.path.exists(path):
        try:
            model = import_module(
                'ollin.estimation.{}.{}'.format(what, name)).Model()
            return model
        except Exception as e:
            print('Unexpected exception occurred while loading model.')
            raise e


def get_estimation_model_list(what):
    path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), what)
    python_files = [
        os.path.basename(module)[:-3]
        for module in glob.glob(os.path.join(path, '*.py'))]
    movement_models = [
            module for module in python_files
            if (
                (module != '__init__') and
                (module != '{}_estimation'.format(what)))]

    print('{} Estimation Model Library:'.format(what.title()))
    for num, mov in enumerate(movement_models):
        print('\t{}.- {}'.format(num + 1, mov))



from ..estimation import Estimate


class OccupancyEstimate(Estimate):
    __slots__ = [
        'occupancy', 'model', 'data', 'detectability']

    def __init__(self, occupancy, model, data, detectability=None):
        self.occupancy = occupancy
        self.model = model
        self.data = data
        self.detectability = detectability

    def __str__(self):
        msg = 'Occupancy estimation done with {} model.\n'
        msg += '\tOccupancy: {}'.format(self.occupancy)
        if self.detectability is not None:
            msg += '\n\tDetectability: {}'.format(self.detectability)
        return msg.format(self.model.name)



from importlib import import_module
import os
import glob

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


@lru_cache()
def load_movement_model(model):
    model_path = os.path.dirname(
            os.path.dirname(os.path.abspath(__file__)))
    model_path = os.path.join(
            model_path, 'movement_models', '{}.py'.format(model))

    if os.path.exists(model_path):
        try:
            cls = import_module(
                    'ollin.movement_models.{}'.format(model)).Model
            return cls
        except Exception as e:
            print('Unexpected exception occured while loading model file')
            raise e


def get_movement_model(model, parameters=None):
    cls = load_movement_model(model)
    return cls(parameters)


def get_movement_model_list():
    """Print all movement models in library."""
    path = os.path.dirname(os.path.abspath(__file__))
    python_files = [
        os.path.basename(module)[:-3]
        for module in glob.glob(os.path.join(path, '*.py'))]
    movement_models = [
        module for module in python_files
        if (module != '__init__') and (module != 'basemodel')]

    print('Movement Model Library:')
    for num, mov in enumerate(movement_models):
        print('\t{}.- {}'.format(num + 1, mov))



"""Constants module for default values.

This module holds all default values for simulator configuration.

Attributes
----------
GLOBAL_CONSTANTS : dict
    This dictionary holds all default values unrelated to movement generation,
    such as the site random generator, default camera configuration
    variables, etc...

MOVEMENT_PARAMETERS : dict
    This dictionary holds default values for any movement model. When extending
    the :py:class:`ollin.movement_models.basemodel.MovementModel` this
    parameters will be added to the model parameters if no replacement is
    provided.

"""

GLOBAL_CONSTANTS = {
    'range': 20,
    'min_clusters': 2,
    'max_clusters': 10,
    'min_cluster_points': 1,
    'max_cluster_points': 10,
    'max_iters': 10,
    'bandwidth_epsilon': 0.01,
    'resolution': 0.1,
    'days': 365,
    'cone_range': 0.01,
    'cone_angle': 60,
    'season': 90,
}

# CONSTANTS FOR MOVEMENT MODELS
MOVEMENT_PARAMETERS = {
    'hr_days': 365,
    'steps_per_day': 4,
    'velocity': {
        'alpha': 0.0,
        'beta': 1.0},
    'home_range': {
        'alpha': 35.0,
        'exponent': 0.54},
    'density': {
        'alpha': 0.0,
        'hr_exp': 1.2,
        'density_exp': 1.75,
        'niche_size_exp': 1.75},
    'movement': {}
}



from six.moves import xrange
import numpy as np
import math
from numba import jit, float64, int64

from .basemodel import MovementModel


class Model(MovementModel):
    name = 'Gradient Levy Model'
    default_parameters = {
        "density": {
            "alpha": 5.5687635749135715,
            "density_exp": 0.9094048898627347,
            "hr_exp": 0.9442292138564827,
            "niche_size_exp": 0.5609397724240494},
        "home_range": {
            "alpha": 102.95793964057235,
            "exponent": 1.8355253220927104},
        "velocity": {
            "alpha": -1.2919142788970375,
            "beta": 2.3185432601405385},
        'movement': {
            'min_pareto': 1.1,
            'max_pareto': 1.9,
            'grad_weight': 8.0},
    }

    def generate_movement(
            self,
            initial_positions,
            site,
            steps,
            velocity):
        min_exponent = self.parameters['movement']['min_pareto']
        max_exponent = self.parameters['movement']['max_pareto']
        grad_weight = self.parameters['movement']['grad_weight']

        heatmap = site.niche
        range_ = site.range
        resolution = site.resolution

        gradient = np.stack(np.gradient(heatmap), -1)

        mov = self._movement(
            gradient,
            heatmap,
            initial_positions,
            resolution,
            velocity,
            range_,
            steps,
            min_exponent,
            max_exponent,
            grad_weight)
        return mov

    @staticmethod
    @jit(
        float64[:, :, :](
            float64[:, :, :],
            float64[:, :],
            float64[:, :],
            float64,
            float64,
            float64[:],
            int64,
            float64,
            float64,
            float64),
        nopython=True)
    def _movement(
            gradient,
            heatmap,
            random_positions,
            resolution,
            velocity,
            range_,
            steps,
            min_exponent,
            max_exponent,
            grad_weight):
        num, _ = random_positions.shape
        movement = np.zeros((num, steps, 2), dtype=float64)
        rangex, rangey = range_
        directions = np.random.uniform(0, 1, size=(steps, num))
        magnitudes = np.random.random((steps, num))
        exponent_var = max_exponent - min_exponent
        gradient = gradient[:, :, 0] + 1j * gradient[:, :, 1]

        for k in xrange(steps):
            movement[:, k, :] = random_positions
            for j in xrange(num):
                direction = directions[k, j]
                magnitude = magnitudes[k, j]
                index = (
                    random_positions[j, 0] // resolution,
                    random_positions[j, 1] // resolution)
                grad = gradient[int(index[0]), int(index[1])]
                value = heatmap[int(index[0]), int(index[1])]

                exponent = min_exponent + exponent_var * value
                magnitude = (velocity * (exponent - 1)) / \
                            (math.pow((1 - magnitude), 1/exponent) * exponent)
                new_angle = (np.angle(grad) +
                             direction / (grad_weight * np.abs(grad) + 1e-10))
                new_direction = (
                    magnitude * math.cos(new_angle),
                    magnitude * math.sin(new_angle))

                tmp1 = (
                    random_positions[j, 0] + new_direction[0],
                    random_positions[j, 1] + new_direction[1])
                tmp2 = (tmp1[0] % (2 * rangex), tmp1[1] % (2 * rangey))

                if tmp2[0] < rangex:
                    random_positions[j, 0] = tmp2[0] % rangex
                else:
                    random_positions[j, 0] = (-tmp2[0]) % rangex

                if tmp2[1] < rangey:
                    random_positions[j, 1] = tmp2[1] % rangey
                else:
                    random_positions[j, 1] = (-tmp2[1]) % rangey
        return movement



from six.moves import xrange
import math
import numpy as np
from numba import jit, float64, int64

from .basemodel import MovementModel


class Model(MovementModel):
    name = 'Variable Levy Model'
    default_parameters = {
        'movement': {
            'min_pareto': 1.1,
            'max_pareto': 1.9},
        "density": {
            "alpha": 5.496771005099745,
            "density_exp": 0.9089427578933271,
            "hr_exp": 0.9313289770069533,
            "niche_size_exp": 0.552259555326986
            },
        "home_range": {
            "alpha": 84.94331014956782,
            "exponent": 1.866467806903153
            },
        "velocity": {
            "alpha": -3.1454639316707027,
            "beta": 7.029550127524596
            }
    }

    def generate_movement(
            self,
            initial_positions,
            site,
            steps,
            velocity):
        min_exponent = self.parameters['movement']['min_pareto']
        max_exponent = self.parameters['movement']['max_pareto']

        heatmap = site.niche
        resolution = site.resolution
        range_ = site.range

        mov = self._movement(
                heatmap,
                initial_positions,
                resolution,
                velocity,
                range_,
                steps,
                min_exponent,
                max_exponent)
        return mov

    @staticmethod
    @jit(
        float64[:, :, :](
            float64[:, :],
            float64[:, :],
            float64,
            float64,
            float64[:],
            int64,
            float64,
            float64),
        nopython=True)
    def _movement(
            heatmap,
            random_positions,
            resolution,
            velocity,
            range_,
            steps,
            min_exponent,
            max_exponent):
        num, _ = random_positions.shape
        movement = np.zeros((num, steps, 2), dtype=float64)
        random_angles = np.random.uniform(0.0, 2 * np.pi, size=(steps, num))
        rangex, rangey = range_
        exponent_var = max_exponent - min_exponent

        for k in xrange(steps):
            movement[:, k, :] = random_positions
            for j in xrange(num):
                angle = random_angles[k, j]
                heading = (math.cos(angle), math.sin(angle))
                index = (
                        random_positions[j, 0] // resolution,
                        random_positions[j, 1] // resolution)
                value = heatmap[int(index[0]), int(index[1])]
                exponent = min_exponent + exponent_var * value
                magnitude = (velocity * (exponent - 1)) / \
                    (math.pow((1 - np.random.rand()), 1/exponent) * exponent)
                direction = (magnitude * heading[0], magnitude * heading[1])
                tmp1 = (
                        random_positions[j, 0] + direction[0],
                        random_positions[j, 1] + direction[1])
                tmp2 = (tmp1[0] % (2 * rangex), tmp1[1] % (2 * rangey))

                if tmp2[0] < rangex:
                    random_positions[j, 0] = tmp2[0] % rangex
                else:
                    random_positions[j, 0] = (-tmp2[0]) % rangex

                if tmp2[1] < rangey:
                    random_positions[j, 1] = tmp2[1] % rangey
                else:
                    random_positions[j, 1] = (-tmp2[1]) % rangey
        return movement



"""Module for generation of camera configuration and detection data.

Cameras can be placed in a virtual world through a
:py:class:`CameraConfiguration`. Given a camera configuration and movement
data, a detection matrix and further detection information can be obtained
through the :py:class:`Detection` class.

Occupancy and other state variables can then be estimated with such detection
data.
"""
from six.moves import xrange
import numpy as np
from scipy.spatial import Voronoi, voronoi_plot_2d

from .constants import GLOBAL_CONSTANTS
from ..estimation.estimation import get_estimation_model


class CameraConfiguration(object):
    """Camera configuration class holding camera positions and directions.

    Attributes
    ----------
    positions : array
        Array of shape [num_cams, 2] to indicate coordinates of each
        camera.
    directions : array
        Array of shape [num_cams, 2] holding a vector of camera direction
        for each camera.
    cone_angle : float
        Viewing angle of cameras in degrees.
    cone_range : float
        Distance to camera at which detection is posible.
    range : array
        Array of shape [2] specifying the dimensions of the virtual world.
    site: :py:obj:`ollin.Site`
        Site object holding information about the virtual world.
    num_cams : int
        Number of cameras.

    """

    def __init__(
            self,
            positions,
            directions,
            site,
            cone_range=None,
            cone_angle=None):
        """Build a camera configuration object.

        Arguments
        ---------
        positions : list or tuple or array
            Array of shape [num_cams, 2] of camera positions.
        directions : list or tuple or array
            Array of shape [num_cams, 2] of camera pointing directions.
        site : :py:obj:`ollin.core.sites.Site`
            Site in which to place cameras.
        cone_range : float, optional
            Distance to camera at which detection is possible. If not provided
            it will be extracted from global constants (see
            :py:const:`ollin.core.constants.GLOBAL_CONSTANTS`).
        cone_angle : float, optional
            Viewing angle of camera in degrees. Default behaviour is as with
            cone_range.

        """
        self.positions = np.array(positions)
        self.directions = np.array(directions)

        self.range = site.range.astype(np.float64)
        self.site = site

        if cone_angle is None:
            cone_angle = GLOBAL_CONSTANTS['cone_angle']
        self.cone_angle = cone_angle

        if cone_range is None:
            cone_range = GLOBAL_CONSTANTS['cone_range']
        self.cone_range = cone_range

        self.num_cams = len(positions)

    def detect(self, mov):
        """Use camera configuration to detect movement history.

        Arguments
        ---------
        mov : :py:obj:`ollin.core.movement.Movement`
            Movement data object to be detected by the camera configuration.

        Returns
        -------
        data : :py:obj:`MovementDetection`
            Camera detection information.

        """
        data = MovementDetection(mov, self)
        return data

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            cone_length=0.4,
            camera_alpha=0.3,
            camera_color=None,
            **kwargs):
        """Draw camera positions and orientations.

        This method will make a plot representing the camera positions and
        orientations in the virtual world. To help visualize the cameras
        corresponding territory Voronoi cells of camera points will be plotted.

        Camera configuration plot adds two components:
            1. "cameras":
                If present in include list, camera position with cone of
                detection will be added to the plot.
            2. "camera_voronoi":
                If present in include list, Voronoi cells for each camera will
                be added to the plot.

        All other components in the include list will be handed down to the
        Site's plotting method. See
        :py:meth:`ollin.core.sites.Site.plot` to
        consult all plotting components defined at that level.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes in which to plot camera info. New axes will be created if none
            are provided.
        figsize : tuple or list, optional
            Size of figure, if ax is not provided. See figsize argument in
            :py:func:`matplotlib.pyplot.figure`.
        include: list or tuple, optional
            List of components to plot. Components list will be passed first
            to the Site object to add the corresponding
            components. Then components corresponding to CameraConfiguration
            included in the list will be plotted.
        cone_length : float, optional
            Length of camera cone for visualization purposes. Defaults to 0.4
            km.
        camera_alpha : float, optional
            Alpha value for camera cones.
        camera_color : str, optional
            Color for camera position points.
        kwargs : dict
            Other keyworded arguments will be passed to the CameraConfiguration
            plot method.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Axes of plot for further processing.

        """
        import matplotlib.pyplot as plt
        from matplotlib.patches import Wedge
        from matplotlib.collections import PatchCollection

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if include is None:
            include = [
                'rectangle',
                'cameras',
                'camera_voronoi']

        ax = self.site.plot(ax=ax, include=include, **kwargs)

        if 'camera_voronoi' in include:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            vor = Voronoi(self.positions)
            voronoi_plot_2d(
                vor, show_points=False, show_vertices=False, ax=ax)
            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)

        if 'cameras' in include:
            if cone_length is None:
                cone_length = self.cone_range

            c_angle = self.cone_angle / 2.0
            patches = []
            for pos, angle in zip(self.positions, self.directions):
                ang = 180 * np.angle(angle[0] + 1j * angle[1]) / np.pi
                wedge = Wedge(pos, cone_length, ang - c_angle, ang + c_angle)
                patches.append(wedge)
            collection = PatchCollection(
                patches, cmap=plt.cm.hsv, alpha=camera_alpha)
            ax.add_collection(collection)

            xcoord, ycoord = self.positions.T
            ax.scatter(xcoord, ycoord, color=camera_color, s=4)

        return ax

    @classmethod
    def make_random(
            cls,
            num,
            site,
            min_distance=None,
            cone_range=None,
            cone_angle=None):
        """Place cameras randomly in range.

        Will create a number of cameras placed at random with random
        directions. If min_distance option is passed, then camera positions
        will be chosen so that any two are not closer that min_distance.

        Arguments
        ---------
        num : int
            Number of cameras to place.
        site : :py:obj:`ollin.core.sites.Site`
            Site in which to place cameras.
        min_distance : float, optional
            Minimum distance in between cameras.
        cone_range : float, optional
            Distance to camera at which detection is possible. If not provided
            it will be extracted from the global constants, see
            :py:const:`ollin.core.constants.GLOBAL_CONSTANTS`.
        cone_angle : float, optional
            Viewing angle of camera in radians. Default behaviour is as with
            cone_range.

        Returns
        -------
        camera : :py:obj:`CameraConfiguration`
            Camera configuration object with random positions and directions.

        """
        range = site.range.astype(np.float64)

        if min_distance is None:
            positions_x = np.random.uniform(0, range[0], size=(num))
            positions_y = np.random.uniform(0, range[1], size=(num))
            positions = np.stack([positions_x, positions_y], -1)
        else:
            positions = make_random_camera_positions(
                num, range, min_distance=min_distance)
        angles = make_random_directions(num)
        cam = cls(
            positions,
            angles,
            site,
            cone_range=cone_range,
            cone_angle=cone_angle)
        return cam

    @classmethod
    def make_grid(
            cls,
            distance,
            site,
            cone_range=None,
            cone_angle=None):
        """Place grid of cameras in virtual world.

        Place cameras in a square grid configuration in range of virtual world,
        parallel to the x and y axis, separated by some distance. Camera
        directions will be random.

        Arguments
        ---------
        distance : float
            Distance between adjacent cameras in grid.
        site : :py:obj:`ollin.core.sites.Site`
            Site in which to place cameras.
        cone_range : float, optional
            Distance to camera at which detection is possible. Default
            behaviour is as with range.
        cone_angle : float, optional
            Viewing angle of camera in radians. Default behaviour is as with
            range.

        Returns
        -------
        camera : :py:obj:`CameraConfiguration`
            Camera configuration object in square grid configuration.

        """
        range = site.range

        num_x = int(range[0] / distance)
        num_y = int(range[1] / distance)

        shift_x = range[0] / num_x
        shift_y = range[1] / num_y

        points_x = np.linspace(0, range[0], num_x, endpoint=False)
        points_y = np.linspace(0, range[1], num_y, endpoint=False)

        X, Y = np.meshgrid(points_x, points_y)
        positions = np.stack((X, Y), -1) + (np.array([shift_x, shift_y]) / 2)
        positions = positions.reshape([-1, 2])
        num = positions.size / 2
        angles = make_random_directions(num)
        cam = cls(
            positions,
            angles,
            site,
            cone_angle=cone_angle,
            cone_range=cone_range)
        return cam


def make_random_camera_positions(num, range, min_distance):
    """Create and return n random points in range separated by min distance.

    Arguments
    ---------
    num : int
        Number of points to create.
    range : tuple or list or array
        Shape of range. range = (a, b) means points will be taken from a a x b
        square.
    min_distance : float
        Distance of minimum separation between points.

    Returns
    -------
    points : array
        Array of shape [num, 2] holding the (x, y) coordinates of points.

    Raises
    ------
    RuntimeError
        If too many cameras are being placed so that minimum distance
        restriction has to be broken.

    """
    random_points_x = np.random.uniform(range[0]/10.0, size=[10, 10, 10])
    random_points_y = np.random.uniform(range[0]/10.0, size=[10, 10, 10])
    random_points = np.stack([random_points_x, random_points_y], -1)
    shift_x = np.linspace(0, range[0], 10, endpoint=False)
    shift_y = np.linspace(0, range[1], 10, endpoint=False)
    shifts = np.stack(np.meshgrid(shift_x, shift_y), -1)
    points = random_points + shifts[:, :, None, :]

    points = points.reshape([-1, 2])
    np.random.shuffle(points)

    selection = [points[0]]
    for i in xrange(num - 1):
        selected = False
        for point in points[1:]:
            is_far = True
            for other_point in selection:
                distance = np.sqrt(
                        (point[0] - other_point[0])**2 +
                        (point[1] - other_point[1])**2)
                if distance <= min_distance:
                    is_far = False
                    break
            if is_far:
                selected = True
                selection.append(point)
                break
        if not selected:
            raise RuntimeError('Cameras dont fit.')

    return np.array(selection)


def make_random_directions(num):
    """Create and return n random direction vectors."""
    angles = np.random.uniform(0, 2*np.pi, size=[num])
    directions = np.stack([np.cos(angles), np.sin(angles)], -1)
    return directions


class Detection(object):
    """Class holding camera detection information.

    Cameras left at site (virtual or real) will make detections at diferent
    steps in time. Which cameras detected when can be stored in a binary
    matrix. This matrix can then be used for estimation of state variables.

    Attributes
    ----------
    camera_config : :py:obj:`CameraConfiguration`
        The camera configuration for the detection data.
    range : array
        Array of shape [2] that holds the dimensions of site.
    detections : array
        Array of shape [steps, num_cams] containing the
        detection information. If::

            detections[j, i] = 1

        then the i-th camera had an detection event at the j-th time step.
    detection_nums : array
        Array of shape [num_cams] with the total number of
        detections for each camera.

    """

    def __init__(self, cam, detections):
        """Construct detection data object.

        Arguments
        ---------
        cam : :py:obj:`CameraConfiguration`
            Camera configuration.
        detections : array
            Array of shape [num_cams, steps] holding the detection
            information.

        """
        self.camera_configuration = cam
        self.range = cam.site.range

        msg = 'Detection array shape implies a different number of cameras'
        msg += ' to number reported from camera configuration'
        assert cam.num_cams == detections.shape[-1], msg

        self.detections = detections
        self.detection_nums = self.detections.sum(axis=0)

    def estimate_occupancy(
            self,
            model='single_species',
            method='MAP',
            priors=None):
        """Estimate occupancy and detectability from detection data.

        Use one of the estimation methods to estimate occupancy and
        detectability, see (:py:mod:`ollin.estimation`)

        Arguments
        ---------
            type : str
                Name of estimation method to use. See
                :py:mod:`ollin.estimation` documentation for a full list.
        Returns
        -------
            estimate : :py:obj:`ollin.estimation.Estimate`
                Estimate object containing estimation information.

        """
        model = get_estimation_model('occupancy', model)
        estimate = model.estimate(self, method=method, priors=priors)
        return estimate

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            detection_cmap='Purples',
            detection_alpha=0.2,
            **kwargs):
        """Plot camera detection data.

        Plots number of detections per camera by coloring the correponding
        Voronoi cell. The number of detections are transformed to [0, 1] scale
        and mapped to a color using a colormap.

        Detection plotting adds the following optional components to the plot:
            1. "detection:
                If present in include list Voronoi regions with color fill,
                encoding the corresponding detection numbers, will be added to
                the plot.
            2. "detection_colorbar":
                If present in include list a colorbar representing the mapping
                between detection numbers and colors will be added to the plot.

        All other components in the include list will be handed down to the
        CameraConfiguration plotting method. See
        :py:meth:`CameraConfiguration.plot` to see all plotting components
        defined there.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot detection information.
        figsize: list or tuple, optional
            Size of figure to create if no axes object was given.
        include: list or tuple, optional
            List of components to plot. Components list will be passed first
            to the Camera Configuration object to add the corresponding
            components. Then components corresponding to CameraConfiguration
            included in the list will be plotted.
        detection_cmap : str, optional
            Colormap with which to encode detection numbers. See
            :py:mod:`matplotlib.cm` to see all options. Defaults to 'Purples'.
        detection_alpha : float, optional
            Alpha value of Voronoi region's color fill.
        kwargs : dict, optional
            Any additional keyworded arguments will be passed to the Camera
            Configuration plot method.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Plot axes for further plotting.

        """
        import matplotlib.pyplot as plt
        from matplotlib.cm import ScalarMappable
        from matplotlib.colors import Normalize

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if include is None:
            include = [
                'rectangle',
                'camera',
                'camera_voronoi',
                'detection',
                'detection_colorbar']

        ax = self.camera_configuration.plot(ax=ax, include=include, **kwargs)

        if 'detection' in include:
            vor = Voronoi(self.camera_configuration.positions)
            cmap = plt.get_cmap(detection_cmap)
            max_num = self.detection_nums.max()
            regions, vertices = voronoi_finite_polygons_2d(vor)
            nums = self.detection_nums
            for reg, num in zip(regions, nums):
                polygon = vertices[reg]
                X, Y = zip(*polygon)
                color = cmap(num / float(max_num))
                ax.fill(X, Y, color=color, alpha=detection_alpha)

            if 'detection_colorbar' in include:
                norm = Normalize(vmin=0, vmax=max_num)
                mappable = ScalarMappable(norm, cmap)
                mappable.set_array(self.detection_nums)
                plt.colorbar(mappable, ax=ax)

        return ax


class MovementDetection(Detection):
    """Class holding detection data arising from movement data.

    From camera placement and movement data, camera detection data can be
    calculated and collected into an array of shape [num_individuals,
    time_steps, num_cameras]. Here::

        array[j, i, k] = 1

    indicates that at the i-th step the j-th
    individual was detected by the k-th camera. Hence more detailed
    analysis is posible from such data.

    Attributes
    ----------
    camera_config : :py:obj:`CameraConfiguration`
        The camera configuration for the detection data.
    range : array
        Array of shape [2] that holds the dimensions of site.
    detections : array
        Array of shape [steps, num_cam] containing the detection information.
        Here::

            detections[j, i] = 1

        means that the i-th camera had an detection event at the j-th time
        step.
    detection_nums : array
        Array of shape [num_cams] with the total number of detections for each
        camera.
    movement : :py:obj:`ollin.core.movement.Movement`
        Movement data being detected.
    grid : array
        Array of shape [num_individuals, time_steps, num_cameras] holding all
        detection data.

    """

    def __init__(self, mov, cam):
        """Construct MovementDetection object.

        Arguments
        ---------
        mov : :py:obj:`ollin.core.movement.Movement`
            Movement data being detected.
        cam : :py:obj:`CameraConfiguration`
            Cameras used for detection.

        """
        msg = "Camera range and movement range do not coincide"
        assert (cam.range == mov.site.range).all(), msg

        self.movement = mov
        self.grid = make_detection_data(mov, cam)
        detections = np.amax(self.grid, axis=0)

        super(MovementDetection, self).__init__(cam, detections)

    def plot(self, ax=None, figsize=(10, 10), include=None, **kwargs):
        """Plot camera detection data.

        Plots number of detections per camera by coloring the correponding
        Voronoi cell. The number of detections are transformed to [0, 1] scale
        and mapped to a color using a colormap.

        Detection plotting adds the following optional components to the plot:
            1. "detection:
                If present in include list Voronoi regions with color fill,
                encoding the corresponding detection numbers, will be added to
                the plot.
            2. "detection_colorbar":
                If present in include list a colorbar representing the mapping
                between detection numbers and colors will be added to the plot.

        All other components in the include list will be passed down to the
        Movement plotting method. See
        :py:meth:`ollin.core.movement.Movement.plot` for all plot
        components defined at that level.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot detection information.
        figsize: list or tuple, optional
            Size of figure to create if no axes object was given.
        include: list or tuple, optional
            List of components to plot. Components list will be passed first
            to the Movement Data object to add the corresponding
            components. Then components corresponding to Movement
            included in the list will be plotted.
        kwargs: dict, optional
            All other keyworded arguments will be passed to Detection and
            Movement plotting methods.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Returns axes for further plotting.

        """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
            ax.set_xticks((0, self.range[0]))
            ax.set_yticks((0, self.range[1]))

        if include is None:
            include = [
                'rectangle',
                'camera',
                'camera_voronoi',
                'detection',
                'detection_colorbar']

        ax = self.movement.plot(ax=ax, include=include, **kwargs)
        ax = super(MovementDetection, self).plot(
            ax=ax, include=include, **kwargs)
        return ax


def make_detection_data(movement, camera_config):
    """Generate and return detection data from movement and camera data.

    Movement data is held in the
    :py:class:`ollin.core.movement.Movement` object and is represented
    by an array of shape [num_individuals, time_steps, 2], which contains the
    full history of movement along the simulated time.

    Use movement history, camera placement and directions to calculate an array
    of shape [num_individuals, time_steps, num_cams] where::

        array[j, k, i] = 1

    implies that the j-th individual was within the detection cone of the i-th
    camera at the k-th time step.

    Arguments
    ---------
    movement : :py:obj:`ollin.core.movement.Movement`
        Movement of individuals to detect.
    camera_config : :py:obj:`CameraConfiguration`
        Camera position and directions to use for detection

    Returns
    -------
    grid : array
        Array of shape [num_individuals, time_steps, num_cameras].

    """
    camera_position = camera_config.positions
    camera_direction = camera_config.directions
    camera_direction = camera_direction[:, 0] + 1j * camera_direction[:, 1]

    movement_data = movement.data
    num, steps, _ = movement_data.shape
    cone_range = camera_config.cone_range
    cone_angle = np.pi * camera_config.cone_angle / 360.0

    relative_pos = (movement_data[:, :, None, :] -
                    camera_position[None, None, :, :])
    relative_pos = relative_pos[:, :, :, 0] + 1j * relative_pos[:, :, :, 1]
    norm = np.abs(relative_pos)
    closeness = np.less(norm, cone_range)

    angles = np.abs(np.angle(relative_pos / camera_direction, deg=1))
    is_in_angle = np.less(angles, cone_angle, where=closeness)

    detected = closeness * is_in_angle
    return detected


def voronoi_finite_polygons_2d(vor, radius=None):
    """Reconstruct infinite voronoi regions in a 2D diagram to finite regions.

    Arguments
    ---------
    vor : Voronoi
        Input diagram
    radius : float, optional
        Distance to 'points at infinity'.

    Returns
    -------
    regions : list of tuples
        Indices of vertices in each revised Voronoi regions.
    vertices : list of tuples
        Coordinates for revised Voronoi vertices. Same as coordinates
        of input vertices, with 'points at infinity' appended to the
        end.

    """
    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge
            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)



from __future__ import division

from ..estimation import EstimationModel
from .occupancy_estimation import OccupancyEstimate


class Model(EstimationModel):
    name = 'Naive Occupancy Estimator'

    def estimate(self, detection, **kwargs):
        steps, cams = detection.detections.shape
        nums = detection.detection_nums

        cams_w_detection = (nums > 0).sum()

        occupancy = cams_w_detection / cams
        detectability = (nums / steps).sum() / cams_w_detection

        est = OccupancyEstimate(
            occupancy, self, detection, detectability=detectability)
        return est



"""Module for Home Range calculation.

Even though home range is a fuzzy concept with several competing definitions,
and estimation methods, in the literature, in this package home range is
defined as:

    **Area occupied by an individual in the course of some fixed time span.**

Home range calculation follows the next pipeline:

1. Inputs are Movement objects that contain position history of all
   simulated individuals for some number of steps
2. Space is discretized using some resolution that only depends on the mean
   velocity of the simulation. See :py:func:`home_range_resolution` to see the
   functional relationship between mean velocity and space resolution
3. For each individual, all pixels in discretized space occupied by the
   individual are counted, and an array with all presence-absence information
   is calculated
4. Each individual is assigned the total area of discretized space occupied
   along its movement

"""

import numpy as np

from .utils import home_range_resolution


class HomeRange(object):
    """Home Range class for storing home range values.

    Movement data is processed into a grid of shape [num_individuals, x, y],
    where x and y are the sizes of discretized space and where::

        grid[i, x, y] = 1

    means that the (x, y) pixel was occupied by the i-th individual.

    Attributes
    ----------
    movement : :py:obj:`ollin.core.movement.Movement`
        Underlying movement data.
    grid : array
        Array of shape [num_individuals, x, y] with prescence-absence
        information.
    resolution : float
        Spatial resolution for space discretization.
    home_ranges : array
        Array of shape [num_individuals] holding the home range in Km^2 for
        each individual.
    mean_home_range : float
        Average home range.

    """

    def __init__(self, movement):
        """Construct Home Range data.

        Arguments
        ---------
        movement : :py:obj:`ollin.core.movement.Movement`
            Movement data for which to calculate home ranges.

        """
        self.movement = movement
        self.resolution = home_range_resolution(movement.velocity)
        self.grid = self._make_grid()
        self.home_ranges = self.grid.sum(axis=(1, 2))
        self.mean_home_range = self.home_ranges.mean()

    def _make_grid(self):
        mov_data = self.movement
        num, steps, _ = mov_data.data.shape
        range_ = mov_data.site.range
        array = mov_data.data

        grid = make_grid(array, range_, self.resolution)
        return grid

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            n_individual=0,
            hr_cmap='Blues',
            **kwargs):
        """Plot home range.

        Plots all pixels occupied by an individual in discretized space.

        Home Range plotting adds the following optional components to the plot:
            1. "home_range:
                If present in include list, all pixels occupied by some subset
                of individuals will be colored. Diferent colors will be chosen
                for different individuals based on some colormap.

        All other components in the include list will be passed down to the
        Movement plotting method. See
        :py:meth:`ollin.core.movement.Movement.plot` for all plot
        components defined at that level.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot detection information.
        figsize : list or tuple, optional
            Size of figure to create if no axes object was given.
        include : list or tuple, optional
            List of components to plot. Components list will be passed first
            to the Movement Data object to add the corresponding
            components. Then components corresponding to Movement
            included in the list will be plotted.
        n_individual : int or list or tuple or array or str, optional
            If int will plot the home range of the corresponding individual.
            If it will plot the home range of all individuals in
            list/tuple/array. Otherwise it must be n_idividual='all', in which
            case it will plot all home ranges.
        hr_cmap : str, optional
            Colormap to use for home ranges, see :py:mod:`matplotlib.cm` to
            see all options. Defaults to 'Blues'.
        kwargs : dict, optional
            All other keyworded arguments will be passed to the Movement
            plotting method.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Returns axes for further plotting.

        """
        import matplotlib.pyplot as plt  # pylint: disable=import-error
        from matplotlib.colors import ListedColormap
        from matplotlib.cm import get_cmap

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if include is None:
            include = [
                    'niche_boundary',
                    'rectangle',
                    'home_range']
        self.movement.plot(ax=ax, include=include, **kwargs)

        if 'home_range' in include:
            is_list = False
            if n_individual == 'all':
                n_individual = range(self.movement.num)
            if isinstance(n_individual, (list, tuple, np.ndarray)):
                home_range = self.grid[np.array(n_individual)]
                is_list = True
            elif n_individual == 'mean':
                home_range = self.grid.mean(axis=0)
            else:
                home_range = self.grid[n_individual]

            _, sizex, sizey = self.grid.shape
            range_ = self.movement.site.range
            rangex, rangey = np.meshgrid(
                np.linspace(0, range_[0], sizex),
                np.linspace(0, range_[1], sizey))

            cmap = get_cmap(hr_cmap)

            if is_list:
                for hr in home_range:
                    color = cmap(np.random.rand())
                    cMap = ListedColormap([(0.0, 0.0, 0.0, 0.0), color])
                    ax.pcolormesh(rangex, rangey, hr.T, cmap=cMap)
            else:
                ax.pcolormesh(rangex, rangey, home_range.T, cmap=cmap)
        return ax


def make_grid(array, range, resolution):
    """Make and return presence absence grid in discretized space.

    Given an array of shape [num, steps, 2] which encodes individual positions
    at diferent time steps in a rectangular area of dimensions=range, and a
    resolution for space discretization, calculates an array of presence and
    absence data of shape [num, x, y], where::

        array[i, x, y] = 1

    means that the i-th individual passed though the (x, y) pixel at some
    point.

    Arguments
    ---------
    array : array
        Array of shape [num, steps, 2] of individuals positions.
    range : tuple or list or array
        Two dimensions of rectangular arena.
    resolution : float
        Resolution of space discretization.

    Returns
    -------
    grid : array
        Array of shape [num, x, y] where x ~ range[0] / resolution and
        y ~ range[1] / resolution.

    """
    num_sides_x = int(np.ceil(range[0] / resolution))
    num_sides_y = int(np.ceil(range[1] / resolution))

    num, steps, _ = array.shape

    grid = np.zeros([num, num_sides_x, num_sides_y])
    indices = np.true_divide(array, resolution).astype(np.int)

    nums = np.linspace(
        0, num,
        num * steps,
        endpoint=False).astype(np.int).reshape([-1, 1])
    xcoords, ycoords = np.split(indices.reshape([-1, 2]), 2, -1)

    ycoords = np.minimum(ycoords, num_sides_y - 1)
    xcoords = np.minimum(xcoords, num_sides_x - 1)

    grid[nums, xcoords, ycoords] = resolution ** 2
    return grid



"""Helper functions."""
from __future__ import division

import numpy as np


def sigmoid(x):
    r"""Sigmoid function.

    .. math::

        \sigma(x) = \frac{1}{1 + e^{-x}}

    """
    return 1 / (1 + np.exp(-x))


def logit(y):
    r"""Logit function.

    .. math::

        \mathop{logit}(y) = -log\left(\frac{1 - y}{y}\right)
    """
    return -np.log((1/y) - 1)


def occupancy_resolution(home_range, parameters=None):
    r"""Get discretization resolution from home range.

    Notes
    -----
    This function is used throughout the package to select spatial resolution
    of site discretization for all occupancy calculations. Such resolution is
    dependant on home range information, and possibly some model parameters.

    Currently, the relation assumed is

    .. math::

        resolution = \sqrt{home\_range}

    Arguments
    ---------
    home_range : float
    parameters : dict, optional
        Possible parameters affecting the home_range -> resolution relation.

    Returns
    -------
    resolution: float

    """
    return np.sqrt(home_range)


def home_range_to_velocity(home_range, parameters=None):
    r"""Get mean velocity from home range information.

    Notes
    This function is used throughout the package to translate home range
    information to mean velocity. Each movement model produces a different
    relation between home range and mean velocity, but a single model is
    applied to all movement models. The general relationship assumed is:

    .. math::

        home\_range = \alpha \cdot velocity^{\beta}

    Parameters are then fitted using simulated data.

    Arguments
    ---------
    home_range : float
    parameters : dict
        The parameters dictionary must contain at two keys, "exponent" and
        "alpha".

    Returns
    -------
    velocity : float

    """
    exponent = parameters['exponent']
    alpha = parameters['alpha']
    return np.power(home_range / alpha, 1 / exponent)


def velocity_to_home_range(velocity, parameters=None):
    r"""Get home range from mean velocity information.

    Notes
    -----
    This function is used throughout the package to translate mean velocity
    to home range information. Each movement model produces a different
    relation between home range and mean velocity, but a single model is
    applied to all movement models. The general relationship assumed is:

    .. math::

        home\_range = \alpha \cdot velocity^{\beta}

    Parameters are then fitted using simulated data.

    Arguments
    ---------
    velocity: float
    parameters : dict
        The parameters dictionary must contain at two keys, "exponent" and
        "alpha".

    Returns
    -------
    home_range : float

    """
    exponent = parameters['exponent']
    alpha = parameters['alpha']
    return alpha * velocity**exponent


def occupancy_to_density(
        occupancy,
        home_range_proportion,
        niche_size,
        parameters=None):
    r"""Get estimated density from occupancy information.

    Notes
    -----
    This function is used throughout the package to translate occupancy
    to density information. This relation also depends on the size of
    the home range in relation to the whole site's area, and to the proportion
    of available adequate space, or size of true niche, see
    :py:meth:`ollin.core.sites.BaseSite.get_niche_size`. Each movement model
    produces a different relation between occupancy and density, but a single
    model (or family of functions) is applied to all movement models. The
    general relationship assumed is:

    .. math::

        logit(o) = \alpha+\beta_1 log(d)+\beta_2 log(ns)+\beta_3 log(hrp)

    where :math:`o, d, ns, hrp` stand for occupancy, density, niche size and
    home range proportion, and :math:`logit(x) = - log(-1 + 1/x)`.
    Equivalently:

    .. math::

        o = \frac{\alpha \cdot d^{\beta_1} \cdot ns^{\beta_2} \cdot
        hrp^{\beta_3}}{1 + \alpha \cdot d^{\beta_1} \cdot
        ns^{\beta_2} \cdot hrp^{\beta_3}}

    Parameters are then fitted using simulated data.

    Arguments
    ---------
    occupancy: float
    home_range_proportion : float
        Number in [0, 1] range representing the proportion of home range area
        to total area.
    niche_size : float
        Proportion of total area that is adequate for species.
    parameters : dict
        The parameters dictionary must contain four keys: "alpha", "hr_exp",
        "density_exp" and "niche_size_exp".

    Returns
    -------
    home_range : float

    """
    alpha = parameters['alpha']
    hr_exp = parameters['hr_exp']
    den_exp = parameters['density_exp']
    nsz_exp = parameters['niche_size_exp']

    density = np.exp(
        (logit(occupancy) -
         alpha -
         np.log(home_range_proportion) * hr_exp -
         np.log(niche_size) * nsz_exp) / den_exp)
    return density


def density_to_occupancy(
        density,
        home_range_proportion,
        niche_size,
        parameters=None):
    r"""Get estimated occupancy from density information.

    Notes
    -----
    This function is used throughout the package to translate density to
    occupancy information. This relation also depends on the size of
    the home range in relation to the whole site's area, and to the proportion
    of available adequate space, or size of true niche, see
    :py:meth:`ollin.core.sites.BaseSite.get_niche_size`. Each movement model
    produces a different relation between occupancy and density, but a single
    model (or family of functions) is applied to all movement models. The
    general relationship assumed is:

    .. math::

        logit(o) = \alpha+\beta_1 log(d)+\beta_2 log(ns)+\beta_3 log(hrp)

    where :math:`o, d, ns, hrp` stand for occupancy, density, niche size and
    home range proportion, and :math:`logit(x) = - log(-1 + 1/x)`.
    Equivalently:

    .. math::

        o = \frac{\alpha \cdot d^{\beta_1} \cdot ns^{\beta_2} \cdot
                hrp^{\beta_3}}{1 + \alpha \cdot d^{\beta_1} \cdot
                ns^{\beta_2} \cdot hrp^{\beta_3}}

    Parameters are then fitted using simulated data.

    Arguments
    ---------
    denisty : float
    home_range_proportion : float
        Number in [0, 1] range representing the proportion of home range area
        to total area.
    niche_size : float
        Proportion of total area that is adequate for species.
    parameters : dict
        The parameters dictionary must contain four keys: "alpha", "hr_exp",
        "density_exp" and "niche_size_exp".

    Returns
    -------
    occupancy : float

    """
    alpha = parameters['alpha']
    hr_exp = parameters['hr_exp']
    den_exp = parameters['density_exp']
    nsz_exp = parameters['niche_size_exp']

    occupancy = sigmoid(
        alpha +
        np.log(density) * den_exp +
        np.log(home_range_proportion) * hr_exp +
        np.log(niche_size) * nsz_exp)
    return occupancy


def home_range_resolution(velocity, parameters=None):
    """Get discretization resolution from mean velocity.

    Notes
    -----
    This function is used throughout the package to select spatial resolution
    of site discretization for all home range calculations. Such resolution is
    dependant on mean velocity information, and possibly other model
    parameters.

    Currently, the relation assumed is

    .. math::

        resolution = velocity

    More complicated relations might be used in the future.

    Arguments
    ---------
    velocity : float
    parameters : dict, optional
        Possible parameters affecting the velocity -> resolution relation.

    Returns
    -------
    resolution: float

    """
    return velocity


def velocity_modification(niche_size, parameters):
    """Get velocity modification at niche size.

    Although for many movement models steps sizes are typically drawn
    from a distribution with a known mean, so that mean velocity can
    be controlled, some models include niche information in step size
    selection which could possible perturb mean step size. Therefore velocity
    correction is necessary.

    The correction depends on the overall niche size, which is the proportion
    of total area with an adequate level of niche value (see
    :py:meth:`ollin.core.site.BaseSite.get_niche_size`) and also on the
    movement model used.

    Based on simulated data, the correct parameters for velocity correction can
    be obtained. See :py:mod:`ollin.calibration.velocity`.

    Arguments
    ---------
    niche_size : float
    parameters : dict
        Dictionary holding the movement model parameters for velocity
        correction.

    """
    alpha = parameters['velocity']['alpha']
    beta = parameters['velocity']['beta']
    return beta + alpha * niche_size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ollin",
    version="0.0.1",
    author="CONABIO, Santiago Martínez",
    author_email="santiago.mbal@gmail.com",
    description="Animal Motion Simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mbsantiago/ollin",
    packages=setuptools.find_packages(),
    install_requires=[
        "six",
        "numpy",
        "numba",
        "scipy",
        "pystan",
        "backports.functools_lru_cache"],
    classifiers=(
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research"
    ),
)



from __future__ import print_function
from multiprocessing import Pool

import sys
import numpy as np
import ollin

from ..core.utils import velocity_to_home_range

TRIALS_PER_WORLD = 1000
NUM_WORLDS = 10
VELOCITIES = [0.1, 0.3, 0.5, 0.8, 1.0, 1.4]
NICHE_SIZES = np.linspace(0.2, 0.9, 4)
RANGE = 20


class HomeRangeCalibrator(object):
    def __init__(
            self,
            movement_model,
            velocities=VELOCITIES,
            niche_sizes=NICHE_SIZES,
            trials_per_world=TRIALS_PER_WORLD,
            num_worlds=NUM_WORLDS,
            range=RANGE):

        self.movement_model = movement_model
        self.velocities = velocities
        self.niche_sizes = niche_sizes
        self.trials_per_world = TRIALS_PER_WORLD
        self.num_worlds = NUM_WORLDS
        self.range = range

        self.hr_info = self.calculate_hr_info()

    def calculate_hr_info(self):
        n_vel = len(self.velocities)
        n_nsz = len(self.niche_sizes)
        num = self.trials_per_world
        mov = self.movement_model

        all_info = np.zeros(
            [n_vel, n_nsz, self.num_worlds, self.trials_per_world])
        arguments = [
            Info(mov, vel, nsz, num, self.range)
            for vel in self.velocities
            for nsz in self.niche_sizes
            for k in xrange(self.num_worlds)]

        print('Simulating {} scenarios'.format(len(arguments)))
        pool = Pool()
        try:
            results = pool.map(get_single_hr_info, arguments)
            pool.close()
            pool.join()
        except KeyboardInterrupt:
            pool.terminate()
            sys.exit()
            quit()

        arguments = [
                (i, j, k)
                for i in xrange(n_vel)
                for j in xrange(n_nsz)
                for k in xrange(self.num_worlds)]

        for (i, j, k), res in zip(arguments, results):
            all_info[i, j, k, :] = res

        return all_info

    def plot(self, cmap='Set2', figsize=(10, 10), ax=None, plotfit=True):
        import matplotlib.pyplot as plt
        from matplotlib.cm import get_cmap
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        cmap = get_cmap(cmap)
        max_hrange = self.hr_info.max()
        for n, oc in enumerate(self.niche_sizes):
            color = cmap(float(n) / len(self.niche_sizes))
            data = self.hr_info[:, n, :, :]
            mean = data.mean(axis=(1, 2))
            std = data.std(axis=(1, 2))

            ax.plot(
                self.velocities,
                mean,
                c=color,
                label='Niche size: {}'.format(oc))
            ax.fill_between(
                self.velocities,
                mean - std,
                mean + std,
                color=color,
                alpha=0.6,
                edgecolor='white')

        if plotfit:
            target_hr = velocity_to_home_range(
                np.array(self.velocities),
                parameters=self.movement_model.parameters['home_range'])
            ax.plot(
                self.velocities,
                target_hr,
                color='red',
                label='target')

        ax.set_yticks(np.linspace(0, max_hrange, 20))
        ax.set_xticks(self.velocities)
        ax.set_xlabel('Velocity (Km/day)')
        ax.set_ylabel('Home range (Km^2)')
        title = 'Home Range Calibration\n{}'
        title = title.format(self.movement_model.name)
        ax.set_title(title)
        ax.legend()
        return ax

    def fit(self):
        from sklearn.linear_model import LinearRegression

        exponents = np.zeros(len(self.niche_sizes))
        alphas = np.zeros(len(self.niche_sizes))
        for num, niche_size in enumerate(self.niche_sizes):
            data = self.hr_info[:, num, :, :]
            concat = []

            for k, vel in enumerate(self.velocities):
                hrdata = data[k, :, :].ravel()
                hrdata = np.stack([vel * np.ones_like(hrdata), hrdata], -1)
                concat.append(hrdata)

            data = np.concatenate(concat, 0)
            velocity, home_range = data.T
            model = LinearRegression()
            model.fit(np.log(velocity)[:, None], np.log(home_range))

            exponents[num] = model.coef_[0]
            alphas[num] = np.exp(model.intercept_)

        fit = {
            'alpha': alphas.mean(),
            'exponent': exponents.mean()}
        return fit


class Info(object):
    __slots__ = [
        'movement_model',
        'velocity',
        'niche_size',
        'num',
        'range']

    def __init__(
            self,
            movement_model,
            velocity,
            niche_size,
            num,
            range_):
        self.movement_model = movement_model
        self.velocity = velocity
        self.niche_size = niche_size
        self.num = num
        self.range = range_


def get_single_hr_info(info):
    site = ollin.Site.make_random(info.niche_size, range=info.range)
    mov = ollin.Movement.simulate(
        site,
        num=info.num,
        velocity=info.velocity,
        days=info.movement_model.parameters['hr_days'],
        movement_model=info.movement_model)
    hr = ollin.HomeRange(mov)
    return hr.home_ranges



import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")

from .core.occupancy import Occupancy
from .core.home_range import HomeRange
from .core.sites import Site, BaseSite
from .core.movement import Movement, MovementData, MovementAnalysis
from .core.detection import (Detection,
                             MovementDetection,
                             CameraConfiguration)

from .movement_models.basemodel import MovementModel
from .movement_models import get_movement_model_list, get_movement_model

from .estimation.estimation import (get_estimation_model,
                                    get_estimation_model_list)

__version__ = '0.0.1'



from six.moves import xrange
import numpy as np
from numba import jit, float64, int64

from .basemodel import MovementModel


class Model(MovementModel):
    name = 'Variable Brownian Model'
    default_parameters = {
        'movement': {
            'niche_weight': 0.2},
        "density": {
            "alpha": 4.956627725964207,
            "density_exp": 0.8509470680476414,
            "hr_exp": 0.8449833387462566,
            "niche_size_exp": 0.8752336080881222
            },
        "home_range": {
            "alpha": 57.61382053222021,
            "exponent": 1.9096337585646315
            },
        "velocity": {
            "alpha": 0.14473388093667755,
            "beta": 0.8918520334954713
            }
    }

    def generate_movement(
            self,
            initial_positions,
            site,
            steps,
            velocity):
        niche_weight = self.parameters['movement']['niche_weight']

        heatmap = site.niche
        resolution = site.resolution
        range_ = site.range

        mov = self._movement(
            heatmap,
            initial_positions,
            resolution,
            velocity,
            range_,
            steps,
            niche_weight)
        return mov

    @staticmethod
    @jit(
        float64[:, :, :](
            float64[:, :],
            float64[:, :],
            float64,
            float64,
            float64[:],
            int64,
            float64),
        nopython=True)
    def _movement(
            heatmap,
            random_positions,
            resolution,
            velocity,
            range_,
            steps,
            niche_weight):
        num, _ = random_positions.shape
        movement = np.zeros((num, steps, 2), dtype=float64)
        sigma = velocity / 1.2533141373155003
        rangex, rangey = range_
        directions = np.random.normal(
            0, sigma, size=(steps, num, 2))

        for k in xrange(steps):
            movement[:, k, :] = random_positions
            for j in xrange(num):
                direction = directions[k, j]
                index = (
                    random_positions[j, 0] // resolution,
                    random_positions[j, 1] // resolution)
                value = heatmap[int(index[0]), int(index[1])]
                direction *= 1 + niche_weight * (0.5 - value)
                tmp1 = (
                    random_positions[j, 0] + direction[0],
                    random_positions[j, 1] + direction[1])
                tmp2 = (tmp1[0] % (2 * rangex), tmp1[1] % (2 * rangey))

                if tmp2[0] < rangex:
                    random_positions[j, 0] = tmp2[0] % rangex
                else:
                    random_positions[j, 0] = (-tmp2[0]) % rangex

                if tmp2[1] < rangey:
                    random_positions[j, 1] = tmp2[1] % rangey
                else:
                    random_positions[j, 1] = (-tmp2[1]) % rangey
        return movement



"""Module for virtual world creation.

All simulations take place withing a rectangular region of the plane called a
site. The dimensions of such site are specified in the range variable.

Additionally, since such a site is meant to be the moving field of some
specific species, any site is complemented with niche information. This niche
information is encoded in a scalar field, a function of the two coordinates,
that provides a measure of "adequacy" of position for the species. The function
values should be in the [0, 1] range, taking a value of 1 to mean the highest
level of adequacy. The function is stored as an array representing the
rectangular region at some spatial resolution. The niche information can then
be exploited to guide individuals in their movements.

Sites can be created randomly by placing a random number of cluster points
in range and making a Gaussian kernel density estimation, or by specifying
points at which niche values are known to be high and extrapolating by some
kernel density estimation. This data could possibly arise from ecological and
climatic variables, real telemetry data, or presence/absence data from camera
traps studies.
"""
from __future__ import division

from abc import abstractmethod
import numpy as np
from six.moves import xrange
from scipy.stats import gaussian_kde

from .constants import GLOBAL_CONSTANTS


class BaseSite(object):
    """Base class for all types of site.

    Attributes
    ----------
    range : array
        Array of shape [2], specifying the dimensions of site (in Km).
    niche : array
        Matrix representing the values of adequacy at different points in site.
    niche_size : float
        Proportion of total area adequate for species. This are points at which
        niche value is above some threshold.
    resolution : float
        Spatial resolution (in Km) of niche array. If ``range = (x, y)`` and
        ``niche.shape = [n, m]`` then ``resolution = (x/n + y/m)/2``.

    """

    def __init__(
            self,
            range,
            niche):
        """Construct BaseSite object.

        Arguments
        ---------
        range : int or float or tuple or list or array
            Dimensions of site in Km. If int or float, it will be assumed that
            site is a square.
        niche : array
            Matrix representing the values of adequacy at different points in
            site.

        """
        if isinstance(range, (int, float)):
            range = np.array([range, range])
        elif isinstance(range, (tuple, list)):
            if len(range) == 1:
                range = [range[0], range[0]]
            range = np.array(range)
        self.range = range.astype(np.float64)

        self.niche = niche
        self.niche_size = self.get_niche_size(niche)
        self.resolution = self.get_niche_resolution(niche, range)

    @staticmethod
    def get_true_niche(niche, threshold=0.25):
        """Select cells with good level of niche adequacy."""
        true_niche = niche >= threshold
        return true_niche

    @staticmethod
    def get_niche_size(niche):
        """Calculate proportion of area of adequate space."""
        true_niche = BaseSite.get_true_niche(niche)
        return true_niche.mean()

    @staticmethod
    def get_niche_resolution(niche, range):
        """Get spatial resolution used in niche array."""
        x, y = range
        n, m = niche.shape
        xres = x / n
        yres = y / m
        return (xres + yres)/2

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            niche_cmap='Reds',
            niche_alpha=1.0,
            boundary_color='black',
            **kwargs):
        """Plot BaseSite information.

        Site plotting adds the following optional components to the
        plot:

        1. "rectangle":
            If present in include list, axes will be fitted to Site's range and
            no ticks will be placed.

        2. "niche":
            If present in include list, a heatmap plot of niche's adequacy
            values will be placed within range. A colormap will be used to
            translate adequacy values to colors.

        2. "niche_boundary":
            If present in include list, the boundary defining the true niche
            will be plotted. The true niche is defined as those cells where
            adequacy value is higher than some threshold, see
            :py:meth:`BaseSite.get_true_niche`.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot site information.
        figsize : list or tuple, optional
            Size of figure to create if no axes object was given. Defaults to
            (10, 10).
        include : list or tuple, optional
            List of components to add to the plot.
        niche_cmap : str, optional
            Colormap to use to codify niche adequacy level to color. Defaults
            to 'Reds'.
        niche_alpha: float, optional
            Alpha value of niche cell colors.
        boundary_color: str, optional
            Color of boundary of true niche.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Return axes for further plotting.

        """
        import matplotlib.pyplot as plt

        if include is None:
            include = ['rectangle', 'niche_boundary', 'niche']

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if 'niche_boundary' in include or 'niche' in include:
            sizex, sizey = self.niche.shape
            rangex, rangey = np.meshgrid(
                np.linspace(0, self.range[0], sizex),
                np.linspace(0, self.range[1], sizey))

            if 'niche_boundary' in include:
                zone = self.get_true_niche(self.niche)
                ax.contour(
                    rangex,
                    rangey,
                    zone.T,
                    levels=0.5,
                    colors=boundary_color)

            if 'niche' in include:
                ax.pcolormesh(
                    rangex,
                    rangey,
                    self.niche.T,
                    cmap=niche_cmap,
                    alpha=niche_alpha)

        if 'rectangle' in include:
            ax.set_xticks(np.linspace(0, self.range[0], 2))
            ax.set_yticks(np.linspace(0, self.range[1], 2))

            ax.set_xlim(0, self.range[0])
            ax.set_ylim(0, self.range[1])

        return ax

    @abstractmethod
    def sample(self, num):
        """Sample n random points from site."""
        pass


class Site(BaseSite):
    """Site with niche from gaussian kernel density estimation.

    A simple way of obtaining an estimate of niche adequacy value, or for
    random niche creation, is to select points in space which are
    known to be adequate for the target species. Interpolation to unknown
    points in space can be accomplished with a gaussian kernel density
    estimation. Selecting different bandwidths for the kernel density
    estimation will result in tighter or more diffuse niche values.

    Attributes
    ----------
    points : array
        Array of shape [num_points, 2] containing the coordinates of the points
        used for the kernel density estimation.
    kde_bandwidth : float
        Bandwidth used in the gaussian kernel density estimation.
    kde : :py:obj:`scipy.stats.gaussian_kde`
        Density estimation object.
    range : array
        Array of shape [2], specifying the dimensions of site (in Km).
    niche : array
        Matrix representing the values of adequacy at different points in site.
    niche_size : float
        Proportion of total area adequate for species. This are points at which
        niche value is above some threshold.
    resolution : float
        Spatial resolution used for niche array construction, in Km.

    """

    def __init__(
            self,
            range,
            points,
            resolution=0.1,
            kde_bandwidth=0.3,
            max_niche_value=1):
        """Construct site object.

        Arguments
        ---------
        range : int or float or tuple or list or array
            Dimensions of site in Km. If int or float, it will be assumed that
            site is a square.
        points : array
            Array of shape [num_points, 2] with coordinates of points to use
            for the kernel density estimation.
        resolution : float
            Spatial resolution used for niche array construction, in Km.
        kde_bandwidth : float
            Bandwidth to use in kernel density estimation.
        resolution : float
            Spatial resolution used for niche array construction, in Km.
        kde_bandwidth : float
            Bandwidth to use in kernel density estimation.
        max_niche_value : float, optional
            After niche construction, niche array will be scaled so that its
            maximum value is this.

        """
        self.points = points
        self.kde_bandwidth = kde_bandwidth

        niche, kde = self.make_niche(points, range, kde_bandwidth, resolution)
        self.kde = kde

        niche = max_niche_value * niche / niche.max()
        super(Site, self).__init__(range, niche)

    def sample(self, num):
        """Use kernel density estimation to sample random points form site."""
        points = self.kde.resample(num).T
        points = np.maximum(
            np.minimum(points, self.range),
            [0, 0])
        return points

    @staticmethod
    def make_niche(points, range, kde_bandwidth, resolution=1.0):
        """Make niche array from points."""
        kde = gaussian_kde(points.T, kde_bandwidth)
        niche = Site.make_niche_from_kde(kde, range, resolution=resolution)
        return niche, kde

    @staticmethod
    def make_niche_from_kde(kde, range, resolution=1.0):
        """Make niche array from kernel density estimation."""
        num_sides_x = int(np.ceil(range[0] / float(resolution)))
        num_sides_y = int(np.ceil(range[1] / float(resolution)))

        shift_x = range[0] / (num_sides_x * 2)
        shift_y = range[1] / (num_sides_y * 2)

        ycoords, xcoords = np.meshgrid(
                np.linspace(0, range[1], num_sides_y, endpoint=False),
                np.linspace(0, range[0], num_sides_x, endpoint=False))
        points = np.stack(
                [xcoords.ravel() + shift_x, ycoords.ravel() + shift_y], 0)
        niche = kde(points).reshape([num_sides_x, num_sides_y])
        return niche

    def plot(
            self,
            ax=None,
            figsize=(10, 10),
            include=None,
            points_color='red',
            **kwargs):
        """Plot Site information.

        Site plotting adds the following optional components to the
        plot:

        1. "points":
            If present in include list, points used for kernel density
            estimation will be show in plot.

        All other components in include list will be passed to BaseSite
        plotting method. See :py:meth:`BaseSite.plot` to see all components
        defined at that level.

        Arguments
        ---------
        ax : :py:obj:`matplotlib.axes.Axes`, optional
            Axes object in which to plot site information.
        figsize : list or tuple, optional
            Size of figure to create if no axes object was given. Defaults to
            (10, 10).
        include : list or tuple, optional
            List of components to add to the plot. Components list will be
            passed to the BaseSite plotting method to add the corresponding
            components.
        points_color : str, optional
            Color of points used for kernel density estimation. Defaults to
            'red'.
        **kwargs : dict, optional
            All other keyworded arguments will be passed to BaseSite plotting
            method.

        Returns
        -------
        ax : :py:obj:`matplotlib.axes.Axes`
            Return axes for further plotting.

        """
        import matplotlib.pyplot as plt

        if include is None:
            include = [
                    'niche_boundary',
                    'niche',
                    'rectangle']
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        ax = super(Site, self).plot(ax=ax, include=include, **kwargs)

        if 'points' in include:
            X, Y = self.points.T
            ax.scatter(X, Y, label='KDE Points')

        return ax

    @classmethod
    def make_random(
            cls,
            niche_size,
            resolution=None,
            range=None,
            min_clusters=None,
            max_clusters=None,
            min_cluster_points=None,
            max_cluster_points=None,
            max_niche_value=1):
        """Make random site.

        Process for random site creation follows the next steps:

        1. Select a random number of clusters. The number is selected within
           the [min_clusters, max_clusters] range.
        2. For each cluster select a central point randomly, using a uniform
           distribution, within site's range.
        3. For each cluster, select a random number of cluster points, within
           the range [min_cluster_points, max_cluster_points], around the
           cluster center, using a gaussian distribution with random covariance
           matrix.
        4. Collect all generated points for use in kernel density estimation.
        5. Select kernel density estimation bandwidth so that niche
           size (see :py:meth:`BaseSite.get_niche_size`) is recovered.

        Arguments
        ---------
        niche_size : float
            Number in [0, 1] range representing the desired proportion of
            adequate niche space to total area.
        resolution : float, optional
            Spatial resolution to use for niche creation. If none is given it
            will be taken from the global constants. See
            :py:const:`ollin.core.constants.GLOBAL_CONSTANTS`.
        range : int or float or list or tuple or array, optional
            Size of created site. If int or float it will be assumed that site
            is square. If none is given it will be taken from the global
            constants.
        min_clusters : int, optional
            Minimum number of clusters used in random niche creation. If none
            is given it will be taken from the global constants.
        max_clusters : int, optional
            Maximum number of clusters used in random niche creation. If none
            is given it will be taken from the global constants.
        min_cluster_points : int, optional
            Minimum number points per cluster used in random niche creation. If
            none is given it will be taken from the global constants.
        max_cluster_points : int, optional
            Maximum number points per cluster used in random niche creation. If
            none is given it will be taken from the global constants.
        max_niche_value : float, optional
            Number in [0, 1] range. Final niche value will have this number as
            a maximum value.

        """
        if resolution is None:
            resolution = GLOBAL_CONSTANTS['resolution']
        if range is None:
            range = GLOBAL_CONSTANTS['range']
        if min_clusters is None:
            min_clusters = GLOBAL_CONSTANTS['min_clusters']
        if max_clusters is None:
            max_clusters = GLOBAL_CONSTANTS['max_clusters']
        if min_cluster_points is None:
            min_cluster_points = GLOBAL_CONSTANTS['min_cluster_points']
        if max_cluster_points is None:
            max_cluster_points = GLOBAL_CONSTANTS['max_cluster_points']

        if isinstance(range, (int, float)):
            range = np.array([range, range])
        elif isinstance(range, (tuple, list)):
            if len(range) == 1:
                range = [range[0], range[0]]
            range = np.array(range)

        points = _make_random_points(
            range, min_clusters, max_clusters, min_cluster_points,
            max_cluster_points)

        bandwidth = _select_bandwidth(range, points, niche_size, resolution)
        site = cls(
            range,
            points,
            resolution=resolution,
            kde_bandwidth=bandwidth,
            max_niche_value=max_niche_value)
        return site


def _make_random_points(range, min_clusters, max_clusters, min_cluster_points,
                        max_cluster_points):
    n_clusters = np.random.randint(min_clusters, max_clusters)

    cluster_centers_x = np.random.uniform(0, range[0], size=[n_clusters])
    cluster_centers_y = np.random.uniform(0, range[1], size=[n_clusters])
    cluster_centers = np.stack([cluster_centers_x, cluster_centers_y], -1)

    points = []
    for k in xrange(n_clusters):
        n_neighbors = np.random.randint(
            min_cluster_points, max_cluster_points)
        centered_points = np.random.normal(size=[n_neighbors, 2])
        variances = np.random.normal(size=[2, 2])
        sheared_points = np.tensordot(
            centered_points, variances, (1, 1))
        shifted_points = sheared_points + cluster_centers[k]
        points.append(shifted_points)

    points = np.concatenate(points, 0)
    return points


def _select_bandwidth(range, points, niche_size, resolution):
    max_iters = GLOBAL_CONSTANTS['max_iters']
    epsilon = GLOBAL_CONSTANTS['bandwidth_epsilon']

    max_bw = range.max() / 2
    min_bw = 0.01

    mid_bw = (max_bw + min_bw) / 2

    kde = gaussian_kde(points.T, mid_bw)

    counter = 0
    while True:
        niche = Site.make_niche_from_kde(kde, range, resolution=resolution)
        niche = niche / niche.max()
        calculated_niche = Site.get_niche_size(niche)

        err = abs(calculated_niche - niche_size)
        if err < epsilon:
            break

        elif calculated_niche < niche_size:
            min_bw = mid_bw
            mid_bw = (max_bw + min_bw) / 2
            kde.set_bandwidth(mid_bw)

        else:
            max_bw = mid_bw
            mid_bw = (max_bw + min_bw) / 2
            kde.set_bandwidth(mid_bw)

        counter += 1
        if counter == max_iters:
            break

    return (min_bw + max_bw) / 2
